/**
 * MiniThreadedProgram.java
 * 
 * Time-stamp: <2010-04-06 17:20:27 ftaiani>
 * @author Francois Taiani   <francois.taiani@comp.lancs.ac.uk>
 * $Id$
 * 
 */

package cryptoMultiThread;

import java.security.Security;

import javax.crypto.Cipher;
import javax.crypto.KeyGenerator;
import javax.crypto.SealedObject;
import javax.crypto.SecretKey;

/**
 * @author Francois Taiani   <francois.taiani@comp.lancs.ac.uk>
 */
public class MiniThreadedProgram {

    public static final int MAX_THREAD  = 300 ;
    public static final int LOOP_ROUNDS = 1000 ;

    private int counter1 = MAX_THREAD*LOOP_ROUNDS ;
    private int counter2 = 0 ;
    private Thread[] children = new Thread[MAX_THREAD] ;

    //    synchronized public void moveCounter1ToCounter2(int value) {
    public void someComplexOperation() throws Exception {
        SecretKey secretKey;
        Cipher encrypter, decrypter;

        secretKey = KeyGenerator.getInstance("DES").generateKey();

        encrypter = Cipher.getInstance("DES");
        encrypter.init(Cipher.ENCRYPT_MODE, secretKey);

        decrypter = Cipher.getInstance("DES");
        decrypter.init(Cipher.DECRYPT_MODE, secretKey);

        String name = "Hello, that's a test";
        SealedObject sealed;
        String unsealed = null ;

        for(int i =0 ; i < LOOP_ROUNDS; i++) {
            sealed = new SealedObject(name, encrypter);
            unsealed = (String) sealed.getObject(decrypter);
        }

        System.out.println("Result: "+unsealed);

    } // EndMethod moveCounter1ToCounter2

    public int getCounter1Plus2() {
	return counter2 + counter1;
    } // EndMethod getCounter1Plus2

    public void startProgram() throws Exception {

	// CREATING ALL THREADS
	for(int i=0; i<MAX_THREAD; i++) {
	    children[i] = new SleepingThread(this,i);
	} // EndFor

	// STARTING ALL THREADS
	for(int i=0; i<MAX_THREAD; i++) {
	    children[i].start();
	} // EndFor

	// WAITING FOR ALL THREADS TO DIE
	for(int i=0; i<MAX_THREAD; i++) {
	    children[i].join();
	} // EndFor
	
        someComplexOperation();

	// PRINTING THE FINAL RESULT
	System.out.println("All threads have died");

    } // EndConstructor MiniThreadedProgram

    public static void main(String[] args) throws Exception {
	MiniThreadedProgram newProgram = new MiniThreadedProgram() ;
	newProgram.startProgram();
    } // EndMethod main

} // EndClass MiniThreadedProgram

class SleepingThread extends Thread {

    private MiniThreadedProgram mainProg ;
    private int myID ;


    public SleepingThread(MiniThreadedProgram theMainProgram, int threadID) {
	mainProg = theMainProgram ;
	myID     = threadID ;
    } // EndMethod SleepingThread

    public void run() {
	try {
            sleep(1000);
	    System.out.println("\nThread "+myID+" has finished.");
	} catch (Exception e) {
	    e.printStackTrace() ;
	    System.exit(1);
	} // EndCatch
    } // EndMethod run

} // EndClass SleepingThread
