/**
 * Engine.java
 * 
 * Time-stamp: <2010-04-05 17:18:32 ftaiani>
 * @author Francois Taiani   <francois.taiani@comp.lancs.ac.uk>
 * $Id$
 * 
 */

package simulation.simulator ;

import simulation.universe.* ;
import java.util.* ;

/**
 * @author Francois Taiani   <francois.taiani@comp.lancs.ac.uk>
 */
public class Engine {

    List<Node>  nodeSet   = new Vector<Node>() ;
    Set<Spring> springSet = new HashSet<Spring>() ;
    public int numberOfNodes ;
    /** Size of the simulation */
    float simuX, simuY;
    float springLenth ;

    public Engine(int nbNodes, float someSimuX, float someSimuY) {
        numberOfNodes = nbNodes;
        simuX = someSimuX ;
        simuY = someSimuY ;
        springLenth = simuX / numberOfNodes ;

        Random randGen = new Random();

        for (int i = 0; i < numberOfNodes ; i++) {
            nodeSet.add( new Node( Math.random()*simuX,
                                   Math.random()*simuY ) );
            if (i>0) {
                springSet.add( new Spring(nodeSet.get(i),
                                          nodeSet.get(randGen.nextInt(i)),
                                          springLenth ) );
            }
        } // EndFor

    } // EndConstructor Engine

    public Iterator<Node> getNodeIterator() {
        return nodeSet.iterator();
    } // EndMethod getNodeIterator

    public void updateSimulationStep(double timeIncrement) {
        for (Iterator<Node> iter = nodeSet.iterator(); iter.hasNext() ;) {
            iter.next().force = new Force();
        }
        for (Iterator<Spring> iter = springSet.iterator(); iter.hasNext() ;) {
            iter.next().computeForce();
        }
        for (Iterator<Node> iter = nodeSet.iterator(); iter.hasNext() ;) {
            iter.next().applyForce(timeIncrement);
        }
       
    } // EndMethod updateSimulationStep

    public static void main(String args[]) {
        if (args.length > 0) {
            if (args[0].equals("-visu")) {
                simulation.visu.SimpleVisualiser.main(args);
            }
        }

        Engine myEngine = new Engine(2000,1200,800);
        
        for(int i=0 ; i < 3000 ; i++) {
            myEngine.updateSimulationStep(0.02);
        }

    } // EndMain


} // EndClass Engine
