/* -*-  Mode:Java; c-basic-offset:2; tab-width:2; indent-tabs-mode:1; -*- */

package src;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.StringTokenizer;

import processing.core.PConstants;
import processing.core.PFont;

public class CircleLayout extends Display {

	public float RADIUS = 120;

// 	final double START_ANGLE = 170 ;
// 	final double END_ANGLE   = 370 ;

	public CircleLayout(VisNode visTree, GraphicPanel graphicPanel) {
		super(visTree, graphicPanel,15);
		RADIUS = (float)(graphicPanel.graphFormat.ORIGIN_RADIUS * Scale);
	}

	public VisNode display(VisNode root, int mouseX, int mouseY, VisNode nodeUnderMouse){
		this.graphicPanel.background(graphicPanel.getBackgroundColour());
		if (!this.graphicPanel.animationLock) {
			this.structureAggregation(root);
			this.recursiveRelocation(root, graphicPanel.graphFormat.START_ANGLE, graphicPanel.graphFormat.END_ANGLE);
		}
		return this.recursiveDisplay(root, mouseX, mouseY, nodeUnderMouse);
	}

	public int structureAggregation(VisNode node) {
		if (node.children.size() == 0 || !node.expanded) {
			node.widthFactor = 1;
			return 1;
		}
		else {
			int totalWidth = 0;
			for(VisNode nextNode : node.children) {
				//			for(int i = 0; i < node.children.size(); i++) {
				totalWidth += structureAggregation(nextNode);
			}
			node.widthFactor = totalWidth;
			return totalWidth;
		}
	}

	// [CHANGE] FT05Apr11 merged with TreeLayout

// 	public VisNode clickedNode;

// 	public VisNode getClickedNode(VisNode visNode, int mouseX, int mouseY) {
// 		getClickedNodeRec(visNode, mouseX, mouseY);
// 		VisNode tmp = clickedNode;
// 		clickedNode = null;
// 		return tmp;
// 	}

// 	public void getClickedNodeRec(VisNode visNode, int mouseX, int mouseY) {

// 		// --- 2010 code
// 		int newHalfWidth = HALF_WIDTH;
// 		int newHalfHeight = HALF_HEIGHT;
// 		if (visNode.profile > 0) {
// 			double scaleVar = this.getScaleVar(visNode, this.maxFrequency);
// 			newHalfWidth = (int)(HALF_WIDTH / scaleVar);
// 			newHalfHeight = (int)(HALF_HEIGHT / scaleVar);
// 		}
// 		// ---

// 		boolean contains = contains(visNode.x, visNode.y, mouseX, mouseY, 
// 				newHalfWidth, newHalfHeight);
// 		if (contains) {
// 			this.clickedNode =  visNode;
// 		}
// 		else {
// 			//			for (int i = 0; i < visNode.children.size(); i++) {
// 			//				VisNode childNode = visNode.children.get(i);
// 			for(VisNode childNode : visNode.children) {
// 				getClickedNodeRec(childNode, mouseX, mouseY);
// 			}
// 		}
// 	}


// 	public boolean contains(int x1, int y1, int x2, int y2, int HALF_WIDTH, int HALF_HEIGHT) {

// 		return (x2 >= x1 - HALF_WIDTH && y2 <= y1 + HALF_HEIGHT)
// 		&& (x2 <= x1 + HALF_WIDTH && y2 >= y1 - HALF_HEIGHT);

// 	}


	/**
	 * Calculates the new location of the nodes 
	 * @param node
	 */
	public void recursiveRelocation(VisNode node, double startDegree, double endDegree) {
		if (node.expanded) {

			int total = 0;
			for(VisNode visChild : node.children) {
				int nextWidthFactor = visChild.widthFactor;
				total += nextWidthFactor;
			}

			int    accumulatedFactor = 0;
			double previousEndDegree = startDegree;
			double degreeRange = endDegree - startDegree;
			
			for(VisNode visChild : node.children) {

				int nextWidthFactor = visChild.widthFactor;
				accumulatedFactor += nextWidthFactor;

				double childDegreeRange = degreeRange * ((double)nextWidthFactor / (double)total);
				double degree = previousEndDegree + childDegreeRange / 2;


				float childLocX = node.x + RADIUS * (float)Math.cos(degree / (180 / Math.PI));
				float childLocY = node.y - RADIUS * (float)Math.sin(degree / (180 / Math.PI));

				visChild.setLocation(childLocX, childLocY);
				recursiveRelocation(visChild, previousEndDegree, previousEndDegree + childDegreeRange);	

				previousEndDegree = previousEndDegree + childDegreeRange;
			}
		}
	}

	
	
	public Hashtable<VisNode, float[]> computeMoves(VisNode visRoot, VisNode futureRoot, boolean moveUp) {
		this.structureAggregation  (futureRoot);
		this.recursiveRelocation   (futureRoot, graphicPanel.graphFormat.START_ANGLE, graphicPanel.graphFormat.END_ANGLE);
		return this.prepareMovementRecords(visRoot, futureRoot, moveUp);
	}
		
	/**********************************************
	 * End of animation computation
	 **********************************************/

	protected void updateDimensions() {
		super.updateDimensions();
		this.RADIUS = (float)(graphicPanel.graphFormat.ORIGIN_RADIUS * Scale);
	} // EndMethod updateDimensions

// 	public void zoomin() {
// // 		this.Scale = this.graphicPanel.SCALE;
// // 		if (this.Scale >= 0.1) {
//  			updateZoom();
// // 		}
// 	} // EndMethod zoomin

// 	public void zoomout() {
// // 		this.Scale = this.graphicPanel.SCALE;
// // 		if (this.Scale <= 2.0) {
//  			updateZoom();
// // 		}
// 	}
}
