/* -*-  Mode:Java; c-basic-offset:2; tab-width:2; indent-tabs-mode:1; -*- */

package src;

import java.util.*;

public class DataProcessor {

	public String[] file;
	public Hashtable<String, ArrayList<String>> rawBranches = new Hashtable<String, ArrayList<String>>();
	public int Total = 0;
	public Hashtable<String, Integer> rawBranchCount = new Hashtable<String, Integer>();
	
	public int MinTreeLevel = -1;

	public boolean isMemoryProfile = false; // true if parse file contains memory information

	Hashtable<String, int[]> colorMap;
	PackageNode packageTree;
	ColorDistribution colorDistribution;

	public DataProcessor(String[] file, double precision) {
		this.file = file;
		this.rawFileProcess(this.file,precision);
		
		/*
		System.out.println("++++++++++++++++++++++++");
		for (int i = 0; i < this.rawBranches.size(); i++) {
			ArrayList<String> nextBranch = this.rawBranches.get(i);
			for (int j = 0; j < nextBranch.size(); j++) {
				System.out.println(nextBranch.get(j));
			}
			System.out.println("---------------------");
		}
		System.out.println("++++++++++++++++++++++++");
		*/
		this.colorMap = new Hashtable<String, int[]>();
		packageTree = new PackageNode("Null");
		packageTree.setParent(null);
		this.colorDistribution = new ColorDistribution();
	}

	private int debugGetCountRawBranchCount() {
		int total = 0;
		for(String key: rawBranchCount.keySet()) {
			total += rawBranchCount.get(key);
		}
		return total;
	}

	public void rawFileProcess(String[] theFile, double precision) {

		Random rand = new Random(0); // with seed for repeatable behaviour

		boolean traceSectionStart = false;
		boolean totalReached      = false; // reach the line of total count
		int initIndex = -1;
		String traceID = "";

		// Temporarily holds the whole set of traces in the hprof
		// file. Some of these traces won't appear in the final tallying
		// table and will be discarded
		Hashtable<String, ArrayList<String>> tempRawBranches = new Hashtable<String, ArrayList<String>>();
		
		for (int i = 0; i < theFile.length; i++) {
			if (theFile[i].contains("--------")) {
				traceSectionStart = true;
			}

			if (traceSectionStart) {
				if (theFile[i].contains("TRACE") || 
						theFile[i].contains("CPU SAMPLES BEGIN") ||
						theFile[i].contains("SITES BEGIN") ) {

					if (initIndex == -1) { // No previous trace. This is the first trace of the file
						initIndex = i + 1;
						
						StringTokenizer stID = new StringTokenizer(theFile[i]);
						stID.nextToken();
						String idStr = stID.nextToken();
						idStr = idStr.substring(0, idStr.length() - 1);
						//System.out.println("whats id: " + idStr);
						traceID = idStr;
						
					} else { // End of a trace, possibly start of a new one

						ArrayList<String> branch = new ArrayList<String>();
						// initIndex contains the position where the ending trace started
						for (int j = initIndex; j < i; j++) {
							StringTokenizer st = new StringTokenizer(theFile[j], "(");
							String tmp = st.nextToken();
							// if (tmp.contains("<empty>")) {
							// 	break;
							// }
							// else {
								branch.add(tmp.trim());
							// }
						}
						
						
						// [CHANGE] FT19Mar11 Removing condition to avoid null
						// pointer exception when empty branches appear in actual
						// count

						tempRawBranches.put(traceID, branch);
						
						StringTokenizer stID = new StringTokenizer(theFile[i]);
						stID.nextToken();
						String idStr = stID.nextToken();
						if (!idStr.contains("SAMPLE")) {
							idStr = idStr.substring(0, idStr.length() - 1);
							//System.out.println("whats id: " + idStr);
							traceID = idStr;
						}
						initIndex = i + 1;
					}
				}
			} // end of traces
			
			if (theFile[i].contains("CPU SAMPLES BEGIN")) {
					// [CHANGE] FT19Mar11 Total no longer part of heading
					// 				StringTokenizer st = new StringTokenizer(theFile[i]);
					// 				st.nextToken(); st.nextToken(); st.nextToken();
					// 				st.nextToken(); st.nextToken();
					// 				String total = st.nextToken();
					// 				total = total.substring(0, total.length() - 1);
					// 				this.Total = Integer.parseInt(total);
					// 				System.out.println(Total);
				totalReached = true;
			}
			
			if(	theFile[i].contains("SITES BEGIN")) {
				System.out.println(theFile[i]); // DEBUG [CHANGE] FT19Mar11
				totalReached     = true;
				isMemoryProfile  = true;
			}
				
			if (totalReached &&
					!theFile[i].contains("CPU") &&
					!theFile[i].contains("rank") &&
					!theFile[i].contains("SITES") &&
					!theFile[i].contains("percent")) {
				//				System.out.println(theFile[i]); // DEBUG [CHANGE] FT19Mar11
				StringTokenizer st = new StringTokenizer(theFile[i]);
					
				st.nextToken();st.nextToken();st.nextToken();
					
				String countStr = st.nextToken(); 

				// [CHANGE] FT19Mar11
				if (isMemoryProfile) {
					st.nextToken();st.nextToken();st.nextToken();
				}
				
				String traceStr = st.nextToken();
				//System.out.println(countStr+" "+traceStr); // DEBUG [CHANGE] FT19Mar11
 					
				int count = Integer.parseInt(countStr);
				// [CHANGE] FT19Mar11 to work with heap
				// profiling as total no longer part of heading

				// TEST
				// System.out.println("DEBUG-BEFORE Total:"+this.Total+" DEBUG rawBranchCount: "+debugGetCountRawBranchCount());
				// if (this.Total!=debugGetCountRawBranchCount()) System.exit(1);
				this.Total+=count;
				// To accomodate empty stacks, always reported with the same
				// IDs, and possibly appearing several times.
				if (rawBranchCount.keySet().contains(traceStr))
					rawBranchCount.put(traceStr, count+rawBranchCount.get(traceStr));
				else
					rawBranchCount.put(traceStr, count);
				// System.out.println("DEBUG Total:"+this.Total+" DEBUG rawBranchCount: "+debugGetCountRawBranchCount());
				if (this.Total!=debugGetCountRawBranchCount()) System.exit(1);
			} // EndIf totalReached and not headerline

		} // EndFor loop through file

		// Sampling down traces for scalability

		System.out.println("Total count "+this.Total+".");

		if (precision > 0){
			System.out.println("Sampling down at precision "+precision+"%.");
			int granularity = (int)(this.Total * (precision/100));

			List<String> branchesToPrune = new ArrayList<String>();

			for(String trace: this.rawBranchCount.keySet()) {
				if (rawBranchCount.get(trace)<granularity) branchesToPrune.add(trace);
			}

			int accumulatedCount = 0;

			while(branchesToPrune.size()>0) {
				// pick a random trace among those to prune. Random selection
				// used to preserve weight of frames closer to root.
				String trace = branchesToPrune.get(rand.nextInt(branchesToPrune.size()));
				branchesToPrune.remove(trace); // each trace only considerd once
				// accumulatedCount used to ensure: (i) probability to be
				// retained proportional to actual weight (e.g. 50% if weight
				// is half of granularity); and (ii) overall error accumulated
				// by dropping some traces is negative and below granularity
				// (no gaussian accumulation)
				accumulatedCount += rawBranchCount.get(trace);
				assert(rawBranchCount.get(trace)<granularity);
				if (accumulatedCount>=granularity) {
					// bumping weight to that of granularity
					rawBranchCount.put(trace, granularity);
					accumulatedCount-=granularity;
				} else {
					rawBranchCount.remove(trace);
				}
			} // EndFor for(String key: this.rawBranchCount.keySet())
		} // EndIf if (precision > 0){


		// Only keeping branches that have a count

		for(String key: this.rawBranchCount.keySet()) {
			this.rawBranches.put(key,tempRawBranches.get(key));
		}

		constructLowestLevelPackageSet();

		System.out.println(tempRawBranches.size()+" traces found.");
		System.out.println(this.rawBranchCount.size()+" traces with count found.");
		System.out.println(this.lowestLevelPackageSet.size()+" individual methods found.");

	} // EndConstructor
	
	public Set<String> lowestLevelPackageSet = new HashSet<String>();

	protected void constructLowestLevelPackageSet() {
		for(List<String> branch: rawBranches.values()) {
			lowestLevelPackageSet.addAll(branch);
		}
	}

	public void constructTree(VisNode visRoot, int currentLevel) {
		
		ArrayList<WeightBranch> weightBranches = new ArrayList<WeightBranch>();
		
		// Enumeration<String> keys = this.rawBranchCount.keys();
		// while(keys.hasMoreElements()) {
		// 	String key = keys.nextElement();

		int totalDebug = 0;

		for(String key: this.rawBranchCount.keySet()) { // [CHANGE] FT27Apr11
			int count = this.rawBranchCount.get(key);
			totalDebug+=count;

			ArrayList<String> branch = this.rawBranches.get(key);
			
			ArrayList<String> newBranch = new ArrayList<String>();
			String lastName = "";
			
			for (int i = branch.size() - 1; i >= 0; i--) {
				String nextName = branch.get(i);
				String updatedName = keepTopLevels(nextName, currentLevel);
				if (!updatedName.equals(lastName)) {
					newBranch.add(updatedName);
				}
				lastName = updatedName;
			}
			
			WeightBranch weightBranch = new WeightBranch(newBranch, count);
			weightBranches.add(weightBranch);
		} // EndFor key
		
		System.out.println("Total debug: "+totalDebug);

		this.constructSpecialBranchTree(visRoot, weightBranches);
	}
	
	public String keepTopLevels(String s, int level) {
		
		StringTokenizer stB = new StringTokenizer(s, ".");
		int totalLevel = stB.countTokens();
		String tmp = "";
		if (level < totalLevel) {
			for (int i = 0; i < level - 1; i++) {
				tmp = tmp + stB.nextToken() + ".";
			}
			tmp = tmp + stB.nextToken();
			return tmp;
		} else {
			return s;
		}
	}
	
	public String truncateLowLevels(String s, int level, int MinTreeLevel) {
		int truncateLevel = MinTreeLevel - level;
		
		String ss = null;
		StringTokenizer stA = new StringTokenizer(s, "(");
		if (stA.countTokens() == 2) {
			ss = stA.nextToken();
		} else {
			ss = s;
		}
		
		if (ss != null) {
			StringTokenizer stB = new StringTokenizer(ss, ".");
			int totalLevel = stB.countTokens();
			if (truncateLevel < totalLevel) {
				String tmp = "";
				int keepLevel = totalLevel - truncateLevel;
				for (int i = 0; i < keepLevel - 1; i++) {
					tmp = tmp + stB.nextToken() + ".";
				}
				tmp = tmp + stB.nextToken();
				//System.out.println(s + " " + tmp + " " + truncateLevel);
				return tmp;
			} 
			else {
				String tmp = stB.nextToken();
				// System.out.println(s + " " + tmp + " " + truncateLevel);
				return tmp;
			}
		}
		return "";
	}


	private VisNode containsVisNode(String profile, SortedSet<VisNode> children) {
		if (children != null) {
// 			for (int i = 0; i < children.size(); i++) {
// 				VisNode tmp = children.get(i);
			for(VisNode tmp : children) {
				//System.out.print(" " + tmp);
				if (tmp.packageName.equals(profile)) {
					return tmp;
				}
			}
			//System.out.println();
		} 
		return null;
	}

	private PackageNode containsPackageNode(String name, ArrayList<PackageNode> children) {
		if (children != null) {
			for (int i = 0; i < children.size(); i++) {
				PackageNode tmp = children.get(i);

				if (tmp.name.equals(name)) {
					return tmp;
				}
			}
		} 
		return null;
	}

	public void aggregateFile() {
		if (this.MinTreeLevel == -1) {
			boolean traceSectionStart = false;
			boolean traceSectionStart2 = false;
			for (int i = 0; i < file.length; i++) {
				if (file[i].contains("--------")) {
					traceSectionStart = true;
				}

				if (traceSectionStart && file[i].contains("TRACE")) {
					traceSectionStart2 = true;
				}

				if (file[i].contains("CPU SAMPLES BEGIN") ||
						file[i].contains("SITES BEGIN")) {
					traceSectionStart = false;
				}

				if (traceSectionStart && traceSectionStart2 && !file[i].contains("TRACE")) {
					// find the minimum level
					StringTokenizer firstST = new StringTokenizer(file[i], "(");
					String packageName = firstST.nextToken().trim();
					StringTokenizer st = new StringTokenizer(packageName, ".");
					int numTokens = st.countTokens();
					if (numTokens > this.MinTreeLevel) {
						this.MinTreeLevel = numTokens;
					}
					// construct the package hierachical tree for color allocation
					PackageNode tmpNode = this.packageTree;
					String tmp = "";
					for (int j = 0; j < numTokens; j++) {
						String nextToken = st.nextToken();
						tmp = tmp + nextToken; 

						// add to a tree
						PackageNode foundNode = containsPackageNode(tmp, tmpNode.children);
						if (foundNode == null) {
							PackageNode child = new PackageNode(tmp);
							child.setParent(tmpNode);
							tmpNode.addChild(child);
							tmpNode = child;
						} else {
							tmpNode = foundNode;
						}

						if (j < numTokens - 1) {
							tmp = tmp + ".";
						}
					}
				}
			}
		}
	}

	public void colorAllocation(PackageNode packageNode, 
			int x_before, int x_after,
			int y_before, int y_after,
			int z_before, int z_after) {
		if (packageNode.children.size() > 0) {
			ArrayList<PackageNode> children = packageNode.children;
			ArrayList<int[]> colors = this.colorDistribution.distribute(children.size(), 
					x_before, x_after, 
					y_before, y_after,
					z_before, z_after);
			//System.out.println(x_before + " " + x_after + " " + y_before + " " +
			//		           y_after + " " + z_before + " " + z_after + " size: " + children.size());
			for (int i = 0; i < children.size(); i++) {
				PackageNode nextChild = children.get(i);
				int[] color = colors.get(i);
				colorMap.put(nextChild.name, color);
				//System.out.println(nextChild.name + " " + color[0] + " " + color[1] + " " + color[2]);
			}
			
			for (int i = 0; i < children.size(); i++) {
				PackageNode nextChild = children.get(i);
				int[] color = colors.get(i);
				// RGB are the centre of the color space allocated to this child
				int colorR = color[0];
				int colorG = color[1];
				int colorB = color[2];

				int x_unit = color[3];
				int y_unit = color[4];
				int z_unit = color[5];
				
				colorAllocation(nextChild, colorR - x_unit, colorR + x_unit,
						colorG - y_unit, colorG + y_unit,
						colorB - z_unit, colorB + z_unit);
			}
		} 
	}

	public void constructSpecialBranchTree(VisNode futureRoot, ArrayList<WeightBranch> branches) {
		
		int count = 0 ;

		for (int i = 0; i < branches.size(); i++) {
			WeightBranch weightBranch = branches.get(i);
			//System.out.println("--------new branch ---------");
			VisNode tmp = futureRoot;
			
			
			for (int j = 0; j < weightBranch.branch.size(); j++) {
				String packageName = weightBranch.branch.get(j);	
				
				VisNode foundNode = containsVisNode(packageName, tmp.children);
				
				if (foundNode == null) {
					// add new node
					VisNode newNode = new VisNode(packageName, weightBranch.weight);
					count++;
					newNode.setParent(tmp);
					tmp.addChild(newNode);
					//System.out.println(tmp.packageName + " adds child " + newNode.packageName);
					tmp = newNode;
				} 
				else {
					//System.out.println(tmp.packageName + " contains child " + foundNode.packageName);
					foundNode.profile += weightBranch.weight;
					tmp = foundNode;
				} 
			}
		} // EndFor i

		System.out.println("Nb of VisNodes:"+count);
	} // EndMethod constructSpecialBranchTree

} // EndClass DataProcessor
