/* -*-  Mode:Java; c-basic-offset:2; tab-width:2; indent-tabs-mode:1; -*- */

package src;

import java.util.Hashtable;
import java.util.ArrayList;
import java.util.StringTokenizer;
import java.text.DecimalFormat;

import processing.core.*;

public abstract class Display {
	
	public Display(VisNode visTree, GraphicPanel graphicPanel, int parentSize) {
		this.visTree      = visTree;
		this.graphicPanel = graphicPanel;
		movedNodes     = new ArrayList<VisNode>();
		this.levelTree = new VisNode("Null", -1);
		
		PARENT_WIDTH  = parentSize;
		PARENT_HEIGHT = parentSize;

		updateZoom();
		// --- 2010 code

		for(VisNode nextNode: visTree.children) {
			if (nextNode.profile > this.maxFrequency) {
				this.maxFrequency = nextNode.profile;
			}
		}
		// ---
	}

	public double getScaleVar(VisNode visNode, int maxFrequency) {
		if (true) {
			return (double) Math.sqrt( (1.0*maxFrequency) / (1.0*visNode.profile) );
		} else {
			//return (double) Math.sqrt( (1.0*maxFrequency) / (1.0*visNode.profile) );
			return 1;	
		}
	}
	
	/**
	 * @return the node under the mouse, null othewise
	 */
 	public abstract VisNode display(VisNode node, int mouseX, int mouseY, VisNode nodeUnderMouse);
 	public abstract Hashtable<VisNode, float[]> 
		computeMoves(VisNode visRoot, VisNode futureRoot, boolean moveUp);
	
 	//public abstract VisNode getClickedNode(VisNode visNode, int mouseX, int mouseY);

	//public PFont myFont;

	public int PARENT_WIDTH = 20;
	public int PARENT_HEIGHT = 20;
	
	public double Scale = 1.0;

	public int H_SPACE = 110;
	public int V_SPACE = 80;

	public VisNode visTree;
	public GraphicPanel graphicPanel;

	public ArrayList<VisNode> movedNodes;
	public VisNode levelTree;
	public int WIDTH = 100;
	public int HEIGHT = 60;
	public int HALF_WIDTH = WIDTH / 2;
	public int HALF_HEIGHT = HEIGHT / 2;

	public int RECT_HEIGHT = 13;
	public int RECT_WIDTH = 3;

	public int CIRCLE_ARROW_RADIUS = 10;
	public int maxFrequency = Integer.MIN_VALUE;

	protected void drawArrow(VisNode visChild, 
													 VisNode node) {

		// --- 2010 code
		double childScaleVar = this.getScaleVar(visChild, this.maxFrequency);
		int newChildHalfWidth    = (int)(HALF_WIDTH / childScaleVar);
		int newChildHalfHeight   = (int)(HALF_HEIGHT / childScaleVar);
		int newCircleArrowRadius = 
			graphicPanel.graphFormat.getArrowSize(this.CIRCLE_ARROW_RADIUS, childScaleVar);

		//draw the circle arrow
		int x_arrow = 0;
		int y_arrow = 0;
					
		// connect to right side of the child node
		if (visChild.x <= node.x - WIDTH) {
			x_arrow = (int)visChild.x + newChildHalfWidth;
			y_arrow = (int)visChild.y; 
		}
		// connect to left side of the child node
		else if (visChild.x >= node.x + WIDTH) {
			x_arrow = (int)visChild.x - newChildHalfWidth;
			y_arrow = (int)visChild.y; 
		}
					
		// connect to bottom side of the child node
		else if ( (visChild.x >= node.x - WIDTH || visChild.x <= node.x + WIDTH)
							&& visChild.y <= node.y) {
			x_arrow = (int)visChild.x;
			y_arrow = (int)visChild.y + newChildHalfHeight;
		}
		// connect to top side of the child node
		else if ( (visChild.x >= node.x - WIDTH || visChild.x <= node.x + WIDTH) 
							&& visChild.y >= node.y) {
			x_arrow = (int)visChild.x;
			y_arrow = (int)visChild.y - newChildHalfHeight;
		} 


		// this.graphicPanel.ellipse(x_arrow, y_arrow, newCircleArrowRadius, newCircleArrowRadius);
					
		this.graphicPanel.pushMatrix();
		this.graphicPanel.translate(x_arrow, y_arrow);
					
		float x = x_arrow - node.x;
		float y = y_arrow - node.y;
		float degree = 0;
					
		if (x > 0 && y == 0) {
			degree = (float) (Math.PI /2);
		}
		else if (x < 0 && y == 0) {
			degree = (float) (-Math.PI / 2);
		}
		else if (x == 0 && y > 0) {
			degree = (float) Math.PI;
		}
		else if (x == 0 && y > 0) {
			degree = (float) 0;
		}
		else if (x > 0 && y > 0) {
			degree = (float)Math.atan2(x, y);
			degree = (float) (Math.PI) - degree;
		}
		else if (x > 0 && y < 0) {
			double tmpY = -y;
			degree = (float)Math.atan2(x, tmpY);
		}
		else if (x < 0 && y > 0) {
			double tmpX = -x;
			degree = (float)Math.atan2(tmpX, y);
			degree = (float) (-Math.PI) + degree;
		}
		else if (x < 0 && y < 0) {
			double tmpX = -x;
			double tmpY = -y;
			degree = (float)-Math.atan2(tmpX, tmpY);
		}
					
		
		int arrowColor = graphicPanel.graphFormat.getArrowColour(node.packageName);

		this.graphicPanel.rotate(degree);
		this.graphicPanel.noStroke();
		this.graphicPanel.smooth();
		this.graphicPanel.fill(arrowColor);
		this.graphicPanel.triangle(-newCircleArrowRadius/2, 
															 newCircleArrowRadius, 
															 0, 0, 
															 newCircleArrowRadius/2, 
															 newCircleArrowRadius);
		this.graphicPanel.popMatrix();
					
		this.graphicPanel.stroke(arrowColor);
		// draw the link between a parent and a child
		this.graphicPanel.line(node.x, node.y, x_arrow, y_arrow);
		this.graphicPanel.noSmooth();

	} // EndMethod drawArrow

	public void drawExpandedNode(VisNode node, double scaleVar, boolean zoomText, boolean dimmed) {
		// System.out.println("scaleVar:" + scaleVar);
		// System.out.println("SCALE:" + graphicPanel.SCALE);
		
		if (!node.packageName.equals("Null")) {
			int parent_x = 0;
			int parent_y = 0;
			
			// int newWidth = (int)(this.WIDTH / scaleVar);
			// int newHeight = (int)(this.HEIGHT / scaleVar);
			int newHalfWidth  = (int)(this.WIDTH  / scaleVar / 2.0);
			int newHalfHeight = (int)(this.HEIGHT / scaleVar / 2.0);
			int newWidth  = newHalfWidth  * 2;
			int newHeight = newHalfHeight * 2;
			
			int newParentWidth =  (int)(this.PARENT_WIDTH / scaleVar);
			int newParentHeight = (int)(this.PARENT_HEIGHT / scaleVar);
			
			// System.out.println("scaleVar:"+scaleVar);

			boolean isLowestLevel = this.graphicPanel.isLowestLevel(node.packageName);
			
			this.graphicPanel.rectMode(PConstants.CENTER);
 			int[] nodeColor = this.graphicPanel.dataProcessor.colorMap.get(node.packageName);
			
			graphicPanel.graphFormat.nodeFormat(nodeColor,dimmed);
			
			this.graphicPanel.rect((int)node.x, (int)node.y, newWidth, newHeight);
			this.graphicPanel.strokeWeight (1);

			// draw the parent color
			String parentName = this.truncateLastLevel(node.packageName);
			int[] parentColor = this.graphicPanel.dataProcessor.colorMap.get(parentName);
			
			if (parentColor != null && graphicPanel.graphFormat.doDrawParentColour) {
				this.graphicPanel.fill(parentColor[0], parentColor[1], parentColor[2]);
				parent_x = (int)node.x + (newHalfWidth - newParentWidth) + (newParentWidth / 2) -1;
				parent_y = (int)node.y - (newHalfHeight - newParentHeight) - (newParentHeight / 2) + 1;
				this.graphicPanel.rect(parent_x, parent_y, newParentWidth, newParentHeight);
			}
			
			// draw the node name
			// this.graphicPanel.textFont(myFont);
			this.graphicPanel.textSize((float)(graphicPanel.graphFormat.ORIGIN_FONT_SIZE * Scale));

			if (!isLowestLevel && graphicPanel.graphFormat.doDrawIcon) {
				// draw the expandible icon
				this.graphicPanel.smooth();
				this.graphicPanel.imageMode(this.graphicPanel.CENTER);
				this.graphicPanel.image(this.graphicPanel.ExpandImage,
																parent_x - newParentWidth, 
																parent_y, 
																newParentWidth, 
																newParentHeight);
				this.graphicPanel.noSmooth();
			}
			int linesAvailable = (int)(newHeight / (graphicPanel.graphFormat.ORIGIN_FONT_SIZE * Scale *0.5));
			if (!zoomText) printNodeName(node,zoomText, linesAvailable, dimmed);
		} 
		else { // Root node
			this.graphicPanel.fill(this.graphicPanel.color(0, 0, 0));
			this.graphicPanel.ellipse(node.x, node.y, CIRCLE_ARROW_RADIUS, CIRCLE_ARROW_RADIUS);
		}
	}

	public void printNodeName(VisNode node, boolean zoomText, int linesAvailable, boolean dimmed) {
	
		int[] nodeColor = this.graphicPanel.dataProcessor.colorMap.get(node.packageName);
			
		String finalPackageName = node.packageName ;
		if (this.graphicPanel.isLowestLevel(node.packageName)) {
			// do not draw the expandible icon
			finalPackageName += "()";
		}
		int linesUsed = 0;
		String text = "";

		if (!zoomText) {
			if (linesAvailable<=0) {linesAvailable=1;}
				
			String[] texts = PackageNameFormatter.formatPackageName(finalPackageName, 20, linesAvailable);
			text = texts[0];
			for (int i = 1; i < texts.length; i++) {
				if (texts[i] != null) {
					text = text + "\n" + texts[i];
					linesUsed++;
				}
			}
		} // End if !zoomText
		else {
			linesUsed = 2;
			text = this.graphicPanel.formatWeightInfo(node);

			this.graphicPanel.textSize((float)(graphicPanel.graphFormat.ORIGIN_FONT_SIZE * Scale * 3));
		} // End if !zoomText else

		graphicPanel.graphFormat.textFormat(nodeColor,zoomText,dimmed);
		this.graphicPanel.textAlign(graphicPanel.CENTER,graphicPanel.CENTER);

		this.graphicPanel.text(text, node.x, node.y);

		this.graphicPanel.textSize((float)(graphicPanel.graphFormat.ORIGIN_FONT_SIZE * Scale));
		this.graphicPanel.textAlign(graphicPanel.LEFT,graphicPanel.BASELINE);

	}

	private String truncateLastLevel(String s) {
		StringTokenizer st = new StringTokenizer(s, ".");
		int numTokens = st.countTokens();
		if (numTokens == 1) {
			return st.nextToken();
		}
		else {
			String tmp = "";
			for (int i = 0; i < numTokens - 2; i++) {
				tmp = tmp + st.nextToken() + ".";
			}
			tmp = tmp + st.nextToken();
			return tmp;
		}
	}


	public boolean contains(int x1, int y1, int x2, int y2, int HALF_WIDTH, int HALF_HEIGHT) {

		return (x2 >= x1 - HALF_WIDTH && y2 <= y1 + HALF_HEIGHT)
		       && (x2 <= x1 + HALF_WIDTH && y2 >= y1 - HALF_HEIGHT);
		
	}

	public boolean isVisNodeUnderMouse(VisNode visNode, int mouseX, int mouseY) {

		int newHalfWidth = HALF_WIDTH;
		int newHalfHeight = HALF_HEIGHT;
		if (visNode.profile > 0) {
			double scaleVar = this.getScaleVar(visNode, this.maxFrequency);
			newHalfWidth  = (int)(HALF_WIDTH / scaleVar);
			newHalfHeight = (int)(HALF_HEIGHT / scaleVar);
		}
		return contains((int)visNode.x, (int)visNode.y, mouseX, mouseY, newHalfWidth, newHalfHeight);
	}

	public VisNode getClickedNode(VisNode visNode, int mouseX, int mouseY) {

		if (isVisNodeUnderMouse(visNode,mouseX,mouseY)) {
			return visNode;
		}
		else {
			for(VisNode childNode: visNode.children) {
				VisNode clickedNode = getClickedNode(childNode, mouseX, mouseY);
				if (clickedNode != null) {
					return clickedNode;
				}
			}
		}
		return null;
	}
	
	public VisNode recursiveDisplay(VisNode node, int mouseX, int mouseY, VisNode nodeUnderMouse) {
		nodeUnderMouse = recursiveVisitNodesInternal(node,mouseX,mouseY,nodeUnderMouse,true);
		if (nodeUnderMouse!=null && nodeUnderMouse!=this.graphicPanel.visRoot)
			printNodeName(nodeUnderMouse,true,2,false);
		return nodeUnderMouse;
	}

	public VisNode findNodeUnderMouse(VisNode node, int mouseX, int mouseY) {
		VisNode nodeUnderMouse = recursiveVisitNodesInternal(node,mouseX,mouseY,null,false);
		return nodeUnderMouse;
	}

	public VisNode recursiveVisitNodesInternal(VisNode node, 
																						 int mouseX, 
																						 int mouseY, 
																						 VisNode nodeUnderMouse,
																						 boolean drawGraph) {

		double scaleVar = this.getScaleVar(node, this.maxFrequency);
		
		boolean nodeUnderMouseIsHere = false;

		if (nodeUnderMouse == null) {
			nodeUnderMouseIsHere = isVisNodeUnderMouse(node, mouseX, mouseY);
			nodeUnderMouse = nodeUnderMouseIsHere?node:null;
		}

		if (!drawGraph) {
			node.highlighted = nodeUnderMouseIsHere;
		}

		if (node.expanded && node.visible) {
			for(VisNode visChild : node.children) {
				if (visChild.visible) {
					if (drawGraph) drawArrow(visChild, node);
					VisNode lowerNoderUnderMouse = recursiveVisitNodesInternal(visChild,mouseX,mouseY, nodeUnderMouse,drawGraph);
					if (nodeUnderMouse == null) nodeUnderMouse = lowerNoderUnderMouse ;
				}
			}

			if (drawGraph) {
				boolean dimmed = !node.highlighted && nodeUnderMouse!=null;
				drawExpandedNode(node, scaleVar, node==nodeUnderMouse, dimmed);
			}
		}
		return nodeUnderMouse;
	} // EndMethod recursiveVisitNodesInternal

	protected void updateDimensions() {
		//this.myFont = this.graphicPanel.createFont("FFScala", (int) (graphicPanel.graphFormat.ORIGIN_FONT_SIZE * Scale));
		H_SPACE = (int)((double)graphicPanel.graphFormat.ORIGIN_H_SPACE * Scale);
		V_SPACE = (int)((double)graphicPanel.graphFormat.ORIGIN_V_SPACE * Scale);
		
		WIDTH  = (int)((double)graphicPanel.graphFormat.ORIGIN_WIDTH  * Scale);
		HEIGHT = (int)((double)graphicPanel.graphFormat.ORIGIN_HEIGHT * Scale);
		HALF_WIDTH  = WIDTH  / 2;
		HALF_HEIGHT = HEIGHT / 2;
		
		RECT_HEIGHT   = (int)((double)graphicPanel.graphFormat.ORIGIN_RECT_HEIGHT   * Scale);
		RECT_WIDTH    = (int)((double)graphicPanel.graphFormat.ORIGIN_RECT_WIDTH    * Scale);

		PARENT_WIDTH  = (int)((double)graphicPanel.graphFormat.ORIGIN_PARENT_WIDTH  * Scale);
		PARENT_HEIGHT = (int)((double)graphicPanel.graphFormat.ORIGIN_PARENT_HEIGHT * Scale);

		this.CIRCLE_ARROW_RADIUS = (int)((double)graphicPanel.graphFormat.ORIGIN_CIRCLE_ARROW_RADIUS * Scale);

	} // EndMethod updateDimensions

	public void updateZoom() {
		this.Scale = this.graphicPanel.graphicState.SCALE;
		updateDimensions();
	} // EndMethod updateZoom

	public void movementProcess(ArrayList<VisNode> nextLowBranch, 
			ArrayList<VisNode> nextHighBranch, 
			Hashtable<VisNode, float[]> movementRecords, 
			boolean moveUp) {
		
		int lowIndex = nextLowBranch.size() - 1;
		int highIndex = nextHighBranch.size() - 1;
		boolean wasEqual = false;
		
		while (lowIndex >= 0 && highIndex >= 0) {
			VisNode lowNode    = nextLowBranch.get(lowIndex);
			VisNode highNode   = nextHighBranch.get(highIndex);
			String  lowString  = lowNode.packageName;
			String  highString = highNode.packageName;
			
			if (this.movementMatch(highString, lowString)) {
				lowIndex--;
				wasEqual = true;
				// add movement record
				this.addMovementRecord(lowNode, highNode, movementRecords, moveUp);
			}
			else {
				if (wasEqual) {
					wasEqual = false;
					highIndex--;
				}
			}
		}
	}
 	
	public int leastAbstractMatch(ArrayList<ArrayList<VisNode>> branches) {
		int leastAbstractIndex = 0;
		int abstractCount = Integer.MIN_VALUE;
		for (int i = 0; i < branches.size(); i++) {
			ArrayList<VisNode> nextBranch = branches.get(i);
			int nextCount = abstractCount(nextBranch);
			if (nextCount > abstractCount) {
				abstractCount = nextCount;
				leastAbstractIndex = i;
			}
		}
		return leastAbstractIndex;
	}
	
	public int branchPatternMatch(ArrayList<VisNode> nextLowBranch, ArrayList<VisNode> nextHighBranch) {
		int highIndex = nextHighBranch.size() - 1;
		int lowIndex = nextLowBranch.size() - 1;

		boolean wasEqual = false;

		while (highIndex >= 0 && lowIndex >= 0) {
			VisNode nextHighNode = nextHighBranch.get(highIndex);
			VisNode nextLowNode = nextLowBranch.get(lowIndex);
			String highString = nextHighNode.packageName;
			String lowString = nextLowNode.packageName;
			
			if (this.movementMatch(highString, lowString)) {

				lowIndex--;
				wasEqual = true;
			} else {

				if (wasEqual) { wasEqual = false; highIndex--;} 
				else {return 0;}
			}
		}
		
		if (highIndex == 0 && lowIndex < 0) {
			return 1; // perfect match, highest priority
		} 
		else if (highIndex < 0 && lowIndex >= 0) {
			return 0; 
		}
		else { //if (highIndex >= 0 && lowIndex < 0)
			return 2;
		}
	}

	private boolean movementMatch (String highLevelPackageName, String lowLevelPackageName) {
		return this.graphicPanel.specialMatches(highLevelPackageName, lowLevelPackageName);
	}

	public int abstractCount(ArrayList<VisNode> branch) {
		int abstractCount = 0;
		for (int i = 0; i < branch.size(); i++) {
			String s = branch.get(i).packageName;
			StringTokenizer st = new StringTokenizer(s, ".");
			abstractCount += st.countTokens();
		}
		return abstractCount;
	}

	/**********************************************
	 * Start of animation computation
	 **********************************************/
	
	public void addMovementRecord(VisNode lowNode, VisNode highNode, Hashtable<VisNode, float[]> movementRecords, boolean moveUp) {
		if (!lowNode.packageName.equals("Null")) {
			// movement record
			if (moveUp) {
				float[] movedLoc = new float[2];
				movedLoc[0] = highNode.x;
				movedLoc[1] = highNode.y;
				movementRecords.put(lowNode, movedLoc);
				
			} else {
				float[] movedLoc = new float[2];
				movedLoc[0] = lowNode.x;
				movedLoc[1] = lowNode.y;
				lowNode.setLocation(highNode.x,highNode.y);
				if (!movementRecords.containsKey(lowNode)) {
					movementRecords.put(lowNode, movedLoc);
				}
			}
		}

	}

	protected Hashtable<VisNode, float[]>
		prepareMovementRecords(VisNode visRoot, 
													 VisNode futureRoot, 
													 boolean moveUp) {
		Hashtable<VisNode, float[]> movementRecords = new Hashtable<VisNode, float[]>();
		
		VisNode highLevelTree = null;
		VisNode lowLevelTree  = null;
		if (moveUp) {
			highLevelTree = futureRoot;
			lowLevelTree = visRoot;
		} else {
			highLevelTree = visRoot;
			lowLevelTree = futureRoot;
		}
		
		ArrayList<ArrayList<VisNode>> lowLevelTreeBranches = this.graphicPanel.tree2Branches(lowLevelTree);
		ArrayList<ArrayList<VisNode>> highLevelTreeBranches = this.graphicPanel.tree2Branches(highLevelTree);
	
		for (int i = 0; i < lowLevelTreeBranches.size(); i++) {
			ArrayList<VisNode> nextLowBranch = lowLevelTreeBranches.get(i);

			ArrayList<ArrayList<VisNode>> MatchBranches_1 = new ArrayList<ArrayList<VisNode>>();
			ArrayList<ArrayList<VisNode>> MatchBranches_2 = new ArrayList<ArrayList<VisNode>>();
			
			boolean foundPerfectMatch = false;
			
			for (int j = 0; j < highLevelTreeBranches.size(); j++) {
				ArrayList<VisNode> nextHighBranch = highLevelTreeBranches.get(j);
				
				//System.out.println(nextLowBranch + "\n" + nextHighBranch + "\n\n");
				
				int result = branchPatternMatch(nextLowBranch, nextHighBranch);
				
				// find the perfect match
				if (result == 1) {
					foundPerfectMatch = true;
					MatchBranches_1.add(nextHighBranch);
				}
				else if (result == 2) {
					MatchBranches_2.add(nextHighBranch);
				}
			}
			
			if (!foundPerfectMatch) {
				//System.out.println("whats here " + MatchBranches_2);
				int index = leastAbstractMatch(MatchBranches_2);
				ArrayList<VisNode> nextHighBranch = MatchBranches_2.get(index);
				//System.out.println("selected match 1" + nextHighBranch);
				movementProcess(nextLowBranch, nextHighBranch, movementRecords,  moveUp);
			}
			else {
				//System.out.println("whats here " + MatchBranches_1);
				int index = leastAbstractMatch(MatchBranches_1);
				ArrayList<VisNode> nextHighBranch = MatchBranches_1.get(index);
				//System.out.println("selected match 2" + nextHighBranch);
				movementProcess(nextLowBranch, nextHighBranch, movementRecords,  moveUp);
			}
			//System.out.println("-------------------");
		}
		
		if (moveUp) {
			// do nothing until animation finished
		} else {
			this.graphicPanel.visRoot = futureRoot;
		}
		return movementRecords;
	}

} // EndClass Display
