/* -*-  Mode:Java; c-basic-offset:2; tab-width:2; indent-tabs-mode:1; -*- */

/**
 * The entry point of the visualisation tool
 * @author Shen Lin
 */

package src;

import java.awt.Dimension;
import java.awt.Toolkit;
import java.io.File;
import java.io.IOException;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Enumeration;
import java.util.GregorianCalendar;
import java.util.Hashtable;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.TreeSet;
import java.lang.reflect.Field;

import javax.swing.JFrame;
import javax.swing.JOptionPane;
import java.awt.*;

import processing.core.*;
import processing.pdf.*;

import treetable.TreeTable;

public class GraphicPanel extends PApplet {
	
	public static int Tutorial = 1; // default to No.
	public static String[] commandArgv ; 

	public GraphFormat graphFormat ;

	public int TutorialStage = 0;
	public boolean TutorialFinishDisplayed = false;
	
	public boolean TutorialWrongAction     = false;

	public static boolean altFormat        = false ;
	public static boolean replayMode       = false;
	public static boolean experimentMode   = false;
	public static boolean debugMode        = false;
	public static String  USERNAME         = "shen";
	public static boolean ENABLE_LOG       = experimentMode;
	public static int     span             = 200 ;
	public static LogGenerator LogGenerator;
	
	public GraphicPanelState graphicState ; // = new GraphicPanelState();
	public GraphicPanelState previousGraphicState = null;
	//	public double SCALE = 1;
	
	public Hashtable<String, Boolean> BranchHideRecord = new Hashtable<String, Boolean>();
	
	public static String ProfilingFile ;

	public PImage ExpandImage = loadImage("lib/expand.png");
	
	public Set<Integer> specialBranchIndices;
	public int clickedNodeIndex = 0;
	
	// not in use any more 
	public ArrayList<ArrayList<String>> specialBranches;
	
	public PFont myFont ;
	
	//public Dimension dimension = Toolkit.getDefaultToolkit().getScreenSize();
	
	public int SCREEN_WIDTH ; //  = (int) dimension.getWidth();
	public int SCREEN_HEIGHT; // = (int) dimension.getHeight();
	
	public int TreeLevel = 1;

	public String unitMeasured;
	
	// style parameter

	public int getBackgroundColour() {
		return graphFormat.BACKGROUND_COLOUR;
	} // EndMethod getBackgroundColour

	public int textColour = 250;
	
	Display       displayMechanism; // layout mechanism
	DataProcessor dataProcessor;
	VisNode       visRoot; // tree structure for display
	VisNode       futureRoot; // tree structure after update
	
	// final public int InitX = this.SCREEN_WIDTH  / 2;
	// final public int InitY = this.SCREEN_HEIGHT / 3;
	
	/** interaction control section */
	int leftBorder = Integer.MAX_VALUE;
	int rightBorder = Integer.MIN_VALUE;
	int upBorder = Integer.MAX_VALUE;
	int downBorder = Integer.MIN_VALUE;
	
	boolean animationLock = false;
	
	Hashtable<VisNode, float[]> animationRecords;
	boolean moveUp;
	
	boolean mouseDragged  = false;
	boolean upDownPressed = false; // indicate a user used left-right
																 // arrow key to change granularity of
																 // the data structure
	boolean disableClick  = false; // disable (un)folding if the
																 // previous animation is not finished
// 	float beforeX;
// 	float beforeY;
// 	float afterX;
// 	float afterY;
	
	String[] file;
	
	HScrollBar hScrollBar;
	VScrollBar vScrollBar;
	
	int TmpVisRootX;
	int TmpVisRootY;
	
	int FadeOutTime  = 60;
	int FadeOutCount =  0;
	// int ClickedNodeX;
	// int ClickedNodeY;
	
	int ClickHintFadeOutTime  = 60;
	int ClickHintFadeOutCount =  0;
	
	int KeyHintFadeOutTime  = 60;
	int KeyHintFadeOutCount =  0;
	
	int AnimationSpeedFactor = 35;
		
	boolean AllVisible = false;
	
	boolean firstMousePressedInDraw = true ; // Hack to avoid jump then Frame gets focus

	String whereToSaveNextFrame = null ;

	public GraphicPanel() {
		super();
	  Dimension dimension = Toolkit.getDefaultToolkit().getScreenSize();
		SCREEN_WIDTH  = (int) dimension.getWidth();
		SCREEN_HEIGHT = (int) dimension.getHeight();
		myFont = createFont("FFScala", 16);
	}

	public GraphicPanel(int width, int height, int fontSize, Frame aFrame) {
		super();
		this.frame = aFrame;
		myFont = createFont("FFScala", fontSize);
		SCREEN_WIDTH  = width ;
		SCREEN_HEIGHT = height ;
	} // EndConstructor GraphicPanel(int width, int height)

	public void initTree(int x, int y) {
		visRoot = new VisNode("Null", -1);
		visRoot.setParent(null);
		visRoot.x = x;
		visRoot.y = y;
		visRoot.setLocation(x, y);
	}
	
	public void initFutureTree(float x, float y) {
		this.futureRoot = new VisNode("Null", -1);
		this.futureRoot.setParent(null);
		this.futureRoot.x = x;
		this.futureRoot.y = y;
		this.futureRoot.setLocation(x, y);
	}
	
	/**
	 * This method loads the tree structure to runtime
	 */
	public void loadTree() {
		//initTree(InitX, InitY);
		//initTree(this.SCREEN_WIDTH /2 , this.SCREEN_HEIGHT / 3);
		initTree(0,0);

		file = loadStrings(ProfilingFile);
		
		if (file==null) {
			System.err.println("Could not load file 'simulator.mem.hprof.txt'. File is empty or inexistant");
			OptionParser.printHelp();
			System.exit(1);
		}

		dataProcessor = new DataProcessor(file,options.getPrecision());
		unitMeasured = dataProcessor.isMemoryProfile?"Bytes":"samples";

		if (altFormat)
			graphFormat  = new AlternativeGraphFormat(this,dataProcessor,span);
		else
			graphFormat  = new GraphFormat(this,dataProcessor,span);

		dataProcessor.aggregateFile();
		graphFormat .colorAllocation();
		dataProcessor.constructTree(visRoot, this.TreeLevel);
	}
	
	public void setup() {
		loadTree();
		graphicState = new GraphicPanelState();
		graphicState.distY = -this.SCREEN_HEIGHT / 6.0;
		background(getBackgroundColour());
		
		noStroke();
		textFont(myFont);

		size(this.SCREEN_WIDTH, this.SCREEN_HEIGHT); 

		// default layout mechanism
		this.displayMechanism = new CircleLayout(visRoot, this);
		//frameRate(40);
		redraw();
		noLoop();
	} // EndMethod setup
	

	public void startAnimation() {
		animationLock = true;
		loop();
	}

	public void stopAnimation() {
		animationLock = false;
		noLoop();
	}

	public int mouseXTranslated() {
		return mouseX - (int)graphicState.distX - getWidth()/2;
	}

	public int mouseYTranslated() {
		return mouseY - (int)graphicState.distY - getHeight()/2;
	}

	public void decideIfRedrawIsNeeded() {
		//System.out.println("in draw");
		if(whereToSaveNextFrame!=null) {
			beginRecord(PDF, whereToSaveNextFrame); 
		}
		//background(getBackgroundColour());
		
		if (mousePressed) {
			int deltaX = (int)(mouseX - pmouseX);
			int deltaY = (int)(mouseY - pmouseY);
			if ((deltaX!=0) || (deltaY!=0)) {
				if (!firstMousePressedInDraw) {
					graphicState.distX += deltaX;
					graphicState.distY += deltaY;
					this.mouseDragged = true;
				}
				firstMousePressedInDraw=false;
			}
		}


		graphicState.nodeUnderMouse = 
			displayMechanism.findNodeUnderMouse(visRoot, mouseXTranslated(), mouseYTranslated());

		if(this.graphicState.hasChanged(previousGraphicState)) redraw();

	} // decideIfRedrawIsNeeded

	private DecimalFormat numberFormatterPercent = new DecimalFormat("#,##0.00");
	private DecimalFormat numberFormatterWeight  = new DecimalFormat("#,##0");

	public String formatWeightInfo(VisNode node) {
		float  result = ((float)node.profile / (float)this.dataProcessor.Total) * 100;
		String percentage = numberFormatterPercent.format(result);  
		String rawMeasure = numberFormatterWeight .format(node.profile);
		String maybeBrackets = (this.isLowestLevel(node.packageName)?"()":"");
		return node.packageName + maybeBrackets + "\n" + percentage + "% (" + rawMeasure + " "+this.unitMeasured+")";
	} // EndMethod formatWeightInfo

	public void displayHint(String hintText) {
		if (!options.isHint()) return ;
		textFont(myFont);
		fill(textColour);
		text(hintText, 50, this.getHeight() - 100);
	} // EndMethod displayHint

	protected float moveAccordingToSpeed(float x, float oldx, float target) {
		//	protected int moveAccordingToSpeed(int x, int oldx, int target, int accel) {
		if (x==target) return x;


		int dir = (x < target)?1:-1;
		int speed = (int)(Math.min(Math.abs(x-oldx),Math.abs(target-x))*0.2)+3;
		//		System.out.println("x: "+x+" oldx:"+oldx+" target:"+target+" speed:"+speed);
		x += speed*dir;
		if (x*dir > target*dir) 
			x = target;

		return x;
	}

	protected void performAnimationIfNeeded() {
		if (this.animationRecords == null) return ;

		boolean animationFinished = true;
			
		for(java.util.Map.Entry<VisNode, float[]> nodeAndTarget: this.animationRecords.entrySet()) {

			VisNode nextVisNode = nodeAndTarget.getKey()  ; //visNodes.nextElement();
			float[] movedLoc    = nodeAndTarget.getValue(); // this.animationRecords.get(nextVisNode);

			if (nextVisNode.x != movedLoc[0] || nextVisNode.y != movedLoc[1])
				animationFinished = false;

			nextVisNode.x = moveAccordingToSpeed(nextVisNode.x,nextVisNode.oldx,movedLoc[0]);
			nextVisNode.y = moveAccordingToSpeed(nextVisNode.y,nextVisNode.oldy,movedLoc[1]);

		}
			
		// Check whether the animation terminates. If terminated and if in
		// the case of moving from low-level tree to high-level tree,
		// need to use the high-level tree instead of the low-level tree.
		if (animationFinished) {
			if (this.moveUp) {
				this.visRoot = this.futureRoot;
				//this.refreshBranchHiding();
			}
			this.animationRecords = null;
			// unlock the animationLock
			stopAnimation();
		}

	} // EndMethod performAnimationIfNeeded 

	public /* synchronized*/ void draw(){

		VisNode clickedNode;

		synchronized(this) {
			previousGraphicState = graphicState.copy();

			performAnimationIfNeeded();

			pushMatrix();
			translate((int)graphicState.distX + getWidth()/2, (int)graphicState.distY + getHeight()/2);
			//scale(graphicState.SCALE);

			//clickedNode = graphicState.nodeUnderMouse;
			clickedNode = 
				this.displayMechanism.display(visRoot, mouseXTranslated(), mouseYTranslated(),null);
			popMatrix();
		}

				
		// information display for mouse over

		if (clickedNode != null && !clickedNode.packageName.equals("Null") 
				&& clickedNode.visible == true) {
			
			/* Mouse is on a node */
			textFont(myFont);
			fill(textColour);
			text(this.formatWeightInfo(clickedNode) + " \n",
					 50, this.getHeight() - 150);

			if (!LogGenerator.mouseEntered) {
				LogGenerator.log("MouseEnter:" + mouseX + "-" + mouseY + ";" +
												 "ClickedNode:" + clickedNode.packageName + ";" +
												 "Profile:" + clickedNode.profile + ";" +
												 "Branch:" + LogGenerator.getBranchPath(clickedNode));
				LogGenerator.mouseEntered = true;
				
				this.KeyHintFadeOutCount = 0;
			}
			
			if (this.ClickHintFadeOutCount <= this.ClickHintFadeOutTime) {
				// textFont(myFont);
				// fill(textColour);
				// text("Hint: left mouse click to expand the node and " +
				// 		 "right mouse click to aggregate the node", 50, this.getHeight() - 100);
				displayHint("Hint: left mouse click to expand the node and " +
										"right mouse click to aggregate the node");
				this.ClickHintFadeOutCount++;
			}
			
		} else {

			/* Mouse is not on a node */

			if (LogGenerator.mouseEntered) {
				LogGenerator.mouseEntered = false;
				LogGenerator.log("MouseLeave");
				this.ClickHintFadeOutCount = 0;
			}
			if (this.KeyHintFadeOutCount <= this.KeyHintFadeOutTime) {
				displayHint("Hint: Press '<-' or '->' key to expand or aggreagte the whole graph.\n" +
										"Hint: Press '-' or '=' key to scale up or down the graph size.\n" +
										"Hint: Press 's' key to re-initialise the graph.");
				this.KeyHintFadeOutCount++;
			}
		}
		
		if ( this.graphicState.furtherCompression == 1 && this.FadeOutCount <= this.FadeOutTime) {
			printRightTopCorner("No further aggregation to avoid ambiguous branches");
			this.FadeOutCount++;
		}
		if ( this.graphicState.furtherCompression == 2 && this.FadeOutCount <= this.FadeOutTime) {
			printRightTopCorner("No further aggregation because of top level package name");
			this.FadeOutCount++;
		}
		if (this.FadeOutCount > this.FadeOutTime) {
			this.FadeOutCount = 0;
			this.graphicState.furtherCompression = 0;
		}
		
		printTutorialMessage();
	} // EndMethod draw()

	private void printRightTopCorner(String msg) {
		if (!options.isHint()) return;
		textFont(myFont);
		fill(textColour);
		textAlign(RIGHT,TOP);
		text(msg, this.getWidth() - 10, 20);
		textAlign(LEFT ,BASELINE);
	} // EndMethod printRightTopCorner

	private void printTutorialMessage() {
		if (Tutorial == 0 && this.TutorialStage == 0) {
			textFont(myFont);
			fill(textColour);
			text("Welcome to the tutorial of ProfilingVis. You will familiarise" +
					 "with the operations that you can perform in ProfilingVis. \n" +
					 "First, press '=' key to increase the size of the graph.", 
					 50, 40);
		}
		
		if (Tutorial == 0 && this.TutorialStage == 1) {
			textFont(myFont);
			fill(textColour);
			text("Well done. Next you can try to " +
					 "press '-' key to decrease the size of the graph.", 
					 50, 40);
		}
		
		if (Tutorial == 0 && this.TutorialStage == 2) {
			textFont(myFont);
			fill(textColour);
			text("Expand the whole graph " +
					 "press '->' key.", 
					 50, 40);
		}
		
		if (Tutorial == 0 && this.TutorialStage == 3) {
			textFont(myFont);
			fill(textColour);
			text("Aggregate the graph " +
					 "press '<-' key.", 
					 50, 40);
		}
		
		if (Tutorial == 0 && this.TutorialStage == 4) {
			textFont(myFont);
			fill(textColour);
			text("Press down the left button on the mouse, and move " +
					 "the mouse around.", 
					 50, 40);
		}
		
		if (Tutorial == 0 && this.TutorialStage == 5) {
			textFont(myFont);
			fill(textColour);
			text("Mouse left click on a node to expand.", 
					 50, 40);
		}
		
		if (Tutorial == 0 && this.TutorialStage == 6) {
			textFont(myFont);
			fill(textColour);
			text("Mouse right click on a node to aggregate.", 
					 50, 40);
		}
		
		if (Tutorial == 1 && this.TutorialStage == 6 && !TutorialFinishDisplayed) {
			JOptionPane.showMessageDialog(this, "You have successfully completed tutorial!");
			this.setup();
			TutorialFinishDisplayed = true;
		}
		
		if (Tutorial == 0 && this.TutorialWrongAction) {
			getMessage();
			TutorialWrongAction = false;
		}
		if(whereToSaveNextFrame!=null) {
			endRecord();
			whereToSaveNextFrame=null;
		}
	} // EndMethod printTutorialMessage
	
	public void mouseMoved() {
		decideIfRedrawIsNeeded();
	}

	public void mouseDragged() {
		decideIfRedrawIsNeeded();
	}

	public void mousePressed() {
		// System.out.println("in mousePressed");
		// [CHANGE] FT08Apr11: experimental
// 		loop();
// 		beforeX = mouseX;
// 		beforeY = mouseY;
		this.firstMousePressedInDraw = true;
		this.upDownPressed = false;
		
		LogGenerator.log("MousePress:" + mouseX + "-" + mouseY);
		decideIfRedrawIsNeeded();
	}
	
	public boolean isLowestLevel(String packageName) {
// 		boolean isLowestLevel = false;
		
// 		Enumeration<ArrayList<String>> enu = this.dataProcessor.rawBranches.elements();
// 		while (enu.hasMoreElements()) {
// 			ArrayList<String> rawBranch = enu.nextElement();
// 			for (int j = 0; j < rawBranch.size(); j++) {
// 				String rawPackageName = rawBranch.get(j);
// 				if (packageName.equals(rawPackageName)) {
// 					isLowestLevel = true;
// 					break;
// 				}
// 			}
// 		}
// 		return isLowestLevel;
		return dataProcessor.lowestLevelPackageSet.contains(packageName);
	}
	
	public void mouseReleased() {
		
		if (!this.mouseDragged) {
			VisNode clickedNode = 
				this.displayMechanism.getClickedNode(visRoot, mouseXTranslated(), mouseYTranslated());
			
			// check visibility
			if (clickedNode != null &&
					!clickedNode.packageName.equals("Null") && 
					clickedNode.visible == true) {
				
				if (mouseButton == LEFT && !disableClick) {
					if (Tutorial == 1 || TutorialStage == 5) {
						String packageName = clickedNode.packageName;
						
						LogGenerator.log("MouseRelease:" + mouseX + "-" + mouseY + ";LeftClick;" +
								"ClickedNode:" + clickedNode.packageName + ";" +
								"Profile:" + clickedNode.profile + ";" +
								"Branch:" + LogGenerator.getBranchPath(clickedNode));
						
						moveUp = false;
						
						// if child of null, and child hide=true
						// set hide false
						Boolean hide = this.BranchHideRecord.get(clickedNode.packageName);
						
						if (clickedNode.getParent().packageName.equals("Null")
								&& hide != null && hide == true)
						{
							//this.BranchHideRecord.put(clickedNode.packageName, false);
							//this.refreshBranchHiding();
						}
						else {
							// otherwise as usually
							if (!isLowestLevel(packageName)) {
								this.onClickChangeAnimation(clickedNode, false);
								//this.refreshBranchHiding();
							}
						}
						
						if (Tutorial == 0 && TutorialStage == 5) {
							TutorialStage++;
						}
					} else {this.TutorialWrongAction = true;} 
				} // end of if left click
				
				
				else if (mouseButton == RIGHT && !disableClick) {
					if (Tutorial == 1 || TutorialStage == 6) {
						
						LogGenerator.log("MouseRelease:" + mouseX + "-" + mouseY + ";RightClick;" +
								"ClickedNode:" + clickedNode.packageName + ";" +
								"Profile:" + clickedNode.profile + ";" +
								"Branch:" + LogGenerator.getBranchPath(clickedNode));
						
						String packageName = clickedNode.packageName;
						VisNode parentNode = clickedNode.getParent();
						String parentName = parentNode.packageName;
						
						// the node is not the lowest level, can compress it
						if (packageName.contains(".")) {
							StringTokenizer st = new StringTokenizer(packageName, ".");
							int level = st.countTokens();
							int count = 0;
							String newPackageName = "";
							while (count < level - 2) {
								newPackageName = newPackageName + st.nextToken() + ".";
								count++;
							}
							newPackageName = newPackageName + st.nextToken();
							
							st = new StringTokenizer(parentName, ".");
							int parentLevel = st.countTokens();
							if (parentLevel >= level) {
								count = 0;
								String newParentName = "";
								while(count < level - 2) {
									newParentName = newParentName + st.nextToken() + ".";
									count++;
								}
								newParentName = newParentName + st.nextToken();
								// after compress, child cannot be the prefix of parent
								if (!newPackageName.equals(newParentName)) {
									moveUp = true;
									this.onClickChangeAnimation(clickedNode, true);
								} else {
									this.graphicState.furtherCompression = 1;
									// this.ClickedNodeX = clickedNode.x;
									// this.ClickedNodeY = clickedNode.y;
								}
							}
							else {
								moveUp = true;
								int resultCode = this.onClickChangeAnimation(clickedNode, true);
								if (resultCode == 0) {
									this.graphicState.furtherCompression = 1;
									// this.ClickedNodeX = clickedNode.x;
									// this.ClickedNodeY = clickedNode.y;
								}
							}
						}
						// show a warning message
						else {
							
							this.graphicState.furtherCompression = 2;
							// this.ClickedNodeX = clickedNode.x;
							// this.ClickedNodeY = clickedNode.y;
						}
						
						if (Tutorial == 0 && TutorialStage == 6) {
							Tutorial = 1; // Finish tutorial can start experiment.
						}
						// disableClick = true;
					}  else {this.TutorialWrongAction = true;}
				} // if end of right click
			} // clicked on a node
			
		} // EndIf  (!this.mouseDragged)
		else {
			LogGenerator.log("MouseDragged:" + mouseX + "-" + mouseY);
			if (graphicState.distX != 0 || graphicState.distY != 0) {
				if (Tutorial == 0 && TutorialStage == 4) {
					TutorialStage ++;
				}
			}
			this.mouseDragged = false;
		} // mouse dragged
		this.upDownPressed = false;
		decideIfRedrawIsNeeded();
	}
	

	protected synchronized void updateScale(double increment) {
		this.graphicState.changeScale(increment);
		this.displayMechanism.updateZoom();
	} // EndMethod updateScale

	
	/** Switch between different representation modes */
	/*synchronized*/ public void keyPressed() {
		// if (keyCode == LEFT && !animationLock) { 
		if (keyCode == LEFT) {
			if (Tutorial == 1 || TutorialStage == 3) {
				LogGenerator.log("Aggregate");
				
				this.upDownPressed = true;
				moveUp = true;
				wholeAnimation();
				
				if (Tutorial == 0 && TutorialStage == 3) {
					TutorialStage ++;
				}
			} 
			else {getMessage();}
		}
		// if (keyCode == RIGHT && !animationLock) {
		if (keyCode == RIGHT) {
			if (Tutorial == 1 || TutorialStage == 2) {
				LogGenerator.log("Expand");
				
				this.upDownPressed = true;
				moveUp = false;
				wholeAnimation();
				
				if (Tutorial == 0 && TutorialStage == 2) {
					TutorialStage ++;
				}
			}
			else {getMessage();}
		}
		if (key == '1') {
			
			if (! (this.displayMechanism instanceof TreeLayout)) {
				this.futureRoot = this.visRoot.copy();
				this.displayMechanism = new TreeLayout(futureRoot, this);
				this.animationRecords = this.displayMechanism.computeMoves(visRoot, futureRoot, true);
				startAnimation();
			}
			
			LogGenerator.log("TreeLayout");
		}
		
		if (key == '2') {
			
			this.initFutureTree(visRoot.x, visRoot.y);
			if (! (this.displayMechanism instanceof CircleLayout)) {
				this.futureRoot = this.visRoot.copy();
				this.displayMechanism = new CircleLayout(futureRoot, this);
				this.animationRecords = this.displayMechanism.computeMoves(visRoot, futureRoot, true);
				startAnimation();
			}
			
			LogGenerator.log("CircleLayout");
		}
		
		if (key == '=') {
			if (Tutorial == 1 || TutorialStage == 0) {
				if (this.graphicState.SCALE < 3) {
					updateScale(0.1);
					// this.graphicState.changeScale(0.1);
					// this.displayMechanism.updateZoom();
					
					LogGenerator.log("ScaleUp");
				}
				
				if (Tutorial == 0 && TutorialStage == 0) {
					TutorialStage ++;
				}
			} 
			else {getMessage();}
		}
		
		if (key == '-') {
			if (Tutorial == 1 || TutorialStage == 1) {
				if (this.graphicState.SCALE > 0.2) {
					updateScale(-0.1);
					// this.graphicState.changeScale(-0.1);
					// this.displayMechanism.updateZoom();
					
					LogGenerator.log("ScaleDown");
				}
				if (Tutorial == 0 && TutorialStage == 1) {
					TutorialStage ++;
				}
			} 
			else {getMessage();}
		}
		
		if (key == 's') {
			int result = JOptionPane.showConfirmDialog
				(this, "Do you really want to re-initialise?", "Reset", JOptionPane.YES_NO_OPTION);
			if (result == 0) { 
				LogGenerator.log("Reset");
				this.setup();
			}
		} // EndIf (key == 's')

		// Disabled by default: buggy on Mac, tends to freeze

 		if (key == 'p') {
 			this.whereToSaveNextFrame = JOptionPane.showInputDialog
 				(this,"Please enter the file to save to:", "profiviz-screenshot.pdf");
 		} // EndIf Printing
		decideIfRedrawIsNeeded();
	}
	
	String[] tutorialMessagesWhenWrong = 
	{ "Please press '='",
			"Please press '-'",
			"Please press '->'",
			"Please press '<-'",
			"Please drag the mouse to move the graph",
			"Please left click on a node to expand it.",
	"Please right click on a node to aggregate it" };
	
	public void getMessage() {
		javax.swing.JOptionPane.showMessageDialog
		(this, 
				tutorialMessagesWhenWrong[TutorialStage]);
	}
	
	private int onClickChangeAnimation(VisNode clickedNode, boolean upDown) {
		//System.out.println("Next Click");
		getSpecialBranches(clickedNode);
		
		this.specialBranchIndices = new TreeSet<Integer>();
		
		// need to find the index of the cliked node on special branches.
		
		ArrayList<ArrayList<String>> oldBranches = new ArrayList<ArrayList<String>>();
		oldBranches.addAll(tree2StringBranches(this.visRoot, clickedNode));
		
		// check no further collapse
		if (upDown) {
			VisNode parentNode = clickedNode.parent;
			if (parentNode != null) {
// 				for (int i = 0; i < parentNode.children.size(); i++) {
// 					VisNode nextChild = parentNode.children.get(i);
				for(VisNode nextChild : parentNode.children) {
					if (nextChild != clickedNode) {
						String childPrefix = 
							this.dataProcessor.truncateLowLevels(nextChild.packageName, 0, 1);
						String clickedNodePrefix = 
							this.dataProcessor.truncateLowLevels(clickedNode.packageName, 0, 1);
						
						
						if (childPrefix.equals(clickedNodePrefix)) {
							//System.out.println("check " + nextChild.packageName + " " + clickedNode.packageName);
							oldBranches.addAll(tree2StringBranches(this.visRoot, nextChild));
						} 
						else if (this.specialMatches(clickedNodePrefix, childPrefix)) {
							return 0;
						}
					}
				}
			}
		} //EndIf if (upDown)
		
		ArrayList<WeightBranch> newBranches = new ArrayList<WeightBranch>();
		Enumeration<String> keys = this.dataProcessor.rawBranchCount.keys();
		
		// We go through all the branches of the hprof graph (the ones in
		// dataProcessor.rawBranches) and we select the one that match the
		// 'compacted branches' that we have just selected in oldBranches
		// above.

		this.LogGenerator.log("{ ");
		
		while(keys.hasMoreElements()) {
			String key = keys.nextElement();
			ArrayList<String> nextRawBranch = this.dataProcessor.rawBranches.get(key);
			int weight = this.dataProcessor.rawBranchCount.get(key);
			
			// find the matching old branch
			ArrayList<String> oldBranch = null;
			int oldBranchIndex = -1;
			
			ArrayList<ArrayList<String>> matchingPerfectOldBranches = new ArrayList<ArrayList<String>>();
			ArrayList<Integer> matchingPerfectIndices = new ArrayList<Integer>();
			
			ArrayList<ArrayList<String>> matchingOldBranches = 
				new ArrayList<ArrayList<String>>();
			ArrayList<Integer> matchingIndices = new ArrayList<Integer>();
			
			boolean foundPerfectMatch = false;
			
			// Going through all oldBranches to see which one match my
			// rawbranch.  There might be several one. If there is a perfect
			// match (i.e. same length), then I take this one. Otherwise, I
			// take the mostAbstractMatch (see below). In the end actual
			// pick might depend on order in oldBranch list

			for (int j = 0; j < oldBranches.size(); j++) {
				ArrayList<String> nextOldBranch = oldBranches.get(j);
				int result = branchPatternMatches(nextOldBranch, nextRawBranch);
				
				if (result == 1) {
					matchingPerfectOldBranches.add(nextOldBranch);
					matchingPerfectIndices.add(j);
					foundPerfectMatch = true;
				} 
				if (result == 2) {
					matchingOldBranches.add(nextOldBranch);
					matchingIndices.add(j);
				}
			} // EndFor j
			
			if ( !foundPerfectMatch ) {
				int final_index = mostAbstractMatch(matchingOldBranches);
				oldBranch       = matchingOldBranches.get(final_index);
				oldBranchIndex  = matchingIndices    .get(final_index);
			} else {
				int final_index = mostAbstractMatch(matchingPerfectOldBranches);
				oldBranch       = matchingPerfectOldBranches.get(final_index);
				oldBranchIndex  = matchingPerfectIndices    .get(final_index);
			}
			
			// if the old branch is special (i.e. on the pacth of the
			// clicked node), record it after updating (i.e. compacting to
			// right level of abstraction)

			if (specialBranchIndices.contains(oldBranchIndex)) {
				ArrayList<String> newBranch = this.getNewSpecialBranch(oldBranch, nextRawBranch, upDown);
				WeightBranch weightBranch = new WeightBranch(newBranch, weight);
				newBranches.add(weightBranch);
				
				this.LogGenerator.log("Trace : " + key + " " + newBranch);
			} 
			else {
				ArrayList<String> updatedBranch = this.updateBranchWithoutChange(nextRawBranch, oldBranch);
				WeightBranch weightBranch = new WeightBranch(updatedBranch, weight);
				newBranches.add(weightBranch);
				
				this.LogGenerator.log("Trace : " + key + " " + updatedBranch);
			}
		} // end while loop
		
		this.LogGenerator.log("} ");
		
		// construct future tree from the newly processed branches
		this.initFutureTree(this.visRoot.x, this.visRoot.y);
		this.dataProcessor.constructSpecialBranchTree(this.futureRoot, newBranches);
		
		// compute movements
		this.animationRecords = this.displayMechanism.computeMoves(this.visRoot, this.futureRoot, upDown);
		//this.visRoot = this.futureRoot; // temporary testing
		startAnimation();
		
		return 1;
	}
	
	public int mostAbstractMatch(ArrayList<ArrayList<String>> branches) {
		int mostAbstractIndex = 0;
		int abstractCount = Integer.MAX_VALUE;
		for (int i = 0; i < branches.size(); i++) {
			ArrayList<String> nextBranch = branches.get(i);
			int nextCount = abstractCount(nextBranch);
			if (nextCount < abstractCount) {
				abstractCount = nextCount;
				mostAbstractIndex = i;
			}
		}
		return mostAbstractIndex;
	}
	
	public int abstractCount(ArrayList<String> branch) {
		int abstractCount = 0;
		for (int i = 0; i < branch.size(); i++) {
			String s = branch.get(i);
			StringTokenizer st = new StringTokenizer(s, ".");
			abstractCount += st.countTokens();
		}
		return abstractCount;
	}
	
	public int branchPatternMatches(ArrayList<String> oldBranch, ArrayList<String> rawBranch) {
		//System.out.print("match check: " + oldBranch + " " + rawBranch);
		int x = oldBranch.size() - 1;
		int y = rawBranch.size() - 1;
		
		boolean wasEqual = false;
		
		while (x >= 0 && y >= 0) {
			String oldString = oldBranch.get(x);
			String rawString = rawBranch.get(y);
			
			if (specialMatches(oldString, rawString)) {
				y--;
				wasEqual = true;
			} else {
				if (wasEqual) { wasEqual = false; x--;} 
				else {return 0;}
			}
		}
		
		if (x == 0 && y < 0) {
			return 1; // perfect match, highest priority
		} 
		else if (x < 0 && y >= 0) {
			return 0; 
		}
		else { //if (highIndex >= 0 && lowIndex < 0)
			return 2;
		}
	}
	
	public ArrayList<ArrayList<String>> tree2StringBranches (VisNode visNode, VisNode clickedNode) {
		ArrayList<VisNode> leafSet = new ArrayList<VisNode>();
		recursiveGetLeafNodes(visNode, leafSet);
		
		ArrayList<ArrayList<String>> branches = new ArrayList<ArrayList<String>>();
		for (int i = 0; i < leafSet.size(); i++) {
			VisNode leafNode = leafSet.get(i);
			ArrayList<String> branch = new ArrayList<String>();
			VisNode tmp = leafNode;
			boolean isSpecialBranch = false;
			while(tmp.parent != null) {
				if (tmp == clickedNode) {
					isSpecialBranch = true;
				}
				branch.add(tmp.packageName);
				tmp = tmp.parent;
			}
			if (isSpecialBranch) {
				this.specialBranchIndices.add(i);
			}
			branches.add(branch);
		}
		return branches;
	}
	
	/**
	 * @return true if oldSpecialString is a 'package prefix' of
	 * rawString. Package prefix means rawString is of the form
	 * oldSpecialString.X.Y... or is equal to oldSpecialString.
	 */
	public boolean specialMatches (String oldSpecialString, String rawString) {
		if (oldSpecialString.length() > rawString.length()) {
			return false;
		}
		else if (oldSpecialString.length() == rawString.length()) {
			return rawString.equals(oldSpecialString);
		}
		else {
			boolean conditionA = rawString.substring(0, oldSpecialString.length()).equals(oldSpecialString);
			boolean conditionB = rawString.charAt(oldSpecialString.length()) == '.';
			return conditionA && conditionB;
		}
	}
	
	private ArrayList<String> getNewSpecialBranch(ArrayList<String> oldSpecialBranch, ArrayList<String> nextRawBranch, boolean upDown) {
		ArrayList<String> newSpecialBranch = new ArrayList<String>();
		int x = oldSpecialBranch.size() - 1;
		int y = nextRawBranch.size() - 1;
		
		boolean wasEqual = false;
		String lastAddedString = null;
		
		//System.out.println("--------------------");
		while (x >= 0 && y >= 0) {
			String oldString = oldSpecialBranch.get(x);
			String rawString = nextRawBranch.get(y);
			
			//System.out.println("before process " + oldString + " " + rawString);
			
			if (specialMatches(oldString, rawString)) {
				wasEqual = true;
				if (x == oldSpecialBranch.size() - this.clickedNodeIndex) {
					
					if (!upDown) {
						StringTokenizer st = new StringTokenizer(oldString, ".");
						int oldLevel = st.countTokens();
						int newLevel = oldLevel + 1;
						
						st = new StringTokenizer(rawString, ".");
						int count = 0;
						String tmp = "";
						while (count < newLevel - 1) {
							tmp = tmp + st.nextToken() + ".";
							count ++;
						}
						tmp = tmp + st.nextToken();
						
						if (lastAddedString == null) {
							lastAddedString = tmp;
							newSpecialBranch.add(tmp);
							//System.out.println("add " + tmp);
						}
						if (!lastAddedString.equals(tmp)) {
							lastAddedString = tmp;
							newSpecialBranch.add(tmp);
							//System.out.println("add 2 " + tmp);
						}
					} else {
						StringTokenizer st = new StringTokenizer(oldString, ".");
						int oldLevel = st.countTokens();
						int newLevel = oldLevel - 1;
						
						int count = 0;
						String tmp = "";
						while (count < newLevel - 1) {
							tmp = tmp + st.nextToken() + ".";
							count ++;
						}
						tmp = tmp + st.nextToken();
						
						if (lastAddedString == null) {
							lastAddedString = tmp;
							newSpecialBranch.add(tmp);
							//System.out.println("add 3" + tmp);
						}
						//System.out.println(lastAddedString + " " + tmp);
						if (!lastAddedString.equals(tmp)) {
							lastAddedString = tmp;
							newSpecialBranch.add(tmp);
							//System.out.println("add 4" + tmp);
						}
					}
				}
				else if (y == 0) {
					if (upDown) {
						if (lastAddedString == null || !specialMatches(lastAddedString, oldString)) {
							lastAddedString = oldString;
							newSpecialBranch.add(oldString);
							//System.out.println("add " + oldString);	
						}
					}
					else {
						lastAddedString = oldString;
						newSpecialBranch.add(oldString);
						//System.out.println("add " + oldString);
					}
				}
				y--;
			} else {
				if (wasEqual) {
					wasEqual = false;
					if (x != oldSpecialBranch.size() - this.clickedNodeIndex) {
						//System.out.println("check " + oldString + " matches last added string " + lastAddedString);
						if (upDown) {
							if (lastAddedString == null || !specialMatches(lastAddedString, oldString)) {
								lastAddedString = oldString;
								newSpecialBranch.add(oldString);
								//System.out.println("add " + oldString);	
							}
						} else {
							lastAddedString = oldString;
							newSpecialBranch.add(oldString);
							//System.out.println("add " + oldString);	
						}
					}
					x--;
				} else {
					return null;
				}
			}
		}
		return newSpecialBranch;
	}
	
	private void getSpecialBranches(VisNode clickedNode) {
		
		this.specialBranches = tree2StringBranches(clickedNode);
		VisNode tmp = clickedNode;
		int parentCount = 0;
		while(tmp.parent != null) {
			parentCount++;
			tmp = tmp.parent;
		}
		
		this.clickedNodeIndex = parentCount;
	}
	
	public ArrayList<ArrayList<String>> exclusiveTree2StringBranches (VisNode visNode, VisNode clickedNode) {
		ArrayList<VisNode> leafSet = new ArrayList<VisNode>();
		recursiveGetLeafNodes(visNode, leafSet);
		
		ArrayList<ArrayList<String>> branches = new ArrayList<ArrayList<String>>();
		for (int i = 0; i < leafSet.size(); i++) {
			boolean exclude = false;
			
			VisNode leafNode = leafSet.get(i);
			ArrayList<String> branch = new ArrayList<String>();
			VisNode tmp = leafNode;
			while(tmp.parent != null) {
				if (tmp == clickedNode) {
					exclude = true;
					break;
				} 
				else {
					branch.add(0, tmp.packageName);
					tmp = tmp.parent;
				}
			}
			if (!exclude) {
				branches.add(branch);
			}
		}
		return branches;
	}
	
	public ArrayList<ArrayList<String>> tree2StringBranches (VisNode visNode) {
		ArrayList<VisNode> leafSet = new ArrayList<VisNode>();
		recursiveGetLeafNodes(visNode, leafSet);
		
		ArrayList<ArrayList<String>> branches = new ArrayList<ArrayList<String>>();
		for (int i = 0; i < leafSet.size(); i++) {
			VisNode leafNode = leafSet.get(i);
			ArrayList<String> reversedBranch = new ArrayList<String>();
			VisNode tmp = leafNode;
			while(tmp.parent != null) {
				reversedBranch.add(tmp.packageName);
				tmp = tmp.parent;
			}
			branches.add(reversedBranch);
		}
		return branches;
	}
	
	public ArrayList<ArrayList<VisNode>> tree2Branches (VisNode visNode) {
		ArrayList<VisNode> leafSet = new ArrayList<VisNode>();
		recursiveGetLeafNodes(visNode, leafSet);
		
		ArrayList<ArrayList<VisNode>> branches = new ArrayList<ArrayList<VisNode>>();
		for (int i = 0; i < leafSet.size(); i++) {
			VisNode leafNode = leafSet.get(i);
			ArrayList<VisNode> reversedBranch = new ArrayList<VisNode>();
			VisNode tmp = leafNode;
			while(tmp.parent != null) {
				reversedBranch.add(tmp);
				tmp = tmp.parent;
			}
			reversedBranch.add(visNode);
			branches.add(reversedBranch);
		}
		return branches;
	}
	
	private void recursiveGetLeafNodes(VisNode node, ArrayList<VisNode> leafSet) {
		if (node.children.size() > 0) {
// 			for (int i = 0; i < node.children.size(); i++) {
// 				VisNode child = node.children.get(i);
			for(VisNode child : node.children) {
				recursiveGetLeafNodes(child, leafSet);
			}
		}
		else {
			leafSet.add(node);
		}
	}
	
	// the whole tree changes level
	private void wholeAnimation() {
		this.initFutureTree(this.visRoot.x, this.visRoot.y);
		
		if (moveUp) {
			// convert the current tree to raw branches
			ArrayList<ArrayList<String>> stringBranches = tree2StringBranches(this.visRoot);
			
			ArrayList<WeightBranch> extendedBranches = new ArrayList<WeightBranch>();
			
			// for each raw string
			Enumeration<String> keys = this.dataProcessor.rawBranchCount.keys(); // traces from hprof file
			
			this.LogGenerator.log("{ ");
			
			while(keys.hasMoreElements()) {
				String key = keys.nextElement();
				ArrayList<String> nextRawBranch = this.dataProcessor.rawBranches.get(key);
				int weight = this.dataProcessor.rawBranchCount.get(key);
				
				// find the matching existing branch
				ArrayList<String> oldBranch = null;
				
				boolean foundPerfectMatch = false;
				
				ArrayList<ArrayList<String>> matchingBranches = new ArrayList<ArrayList<String>>();
				ArrayList<ArrayList<String>> matchingPerfectBranches = new ArrayList<ArrayList<String>>();
				
				
				//System.out.println("RAWBRANCH: nextRawBranch: "   +nextRawBranch);

				for (int j = 0; j < stringBranches.size(); j++) {
					ArrayList<String> nextOldBranch = stringBranches.get(j); // visual tree before expansion
					
					int result = branchPatternMatches(nextOldBranch, nextRawBranch);

					//System.out.println("OLDBRANCH: nextOldBranch "   +nextOldBranch);
					//System.out.println("RESULT: "   +result);

					
					if (result == 1) {
						matchingPerfectBranches.add(nextOldBranch);
						foundPerfectMatch = true;
					}
					if (result == 2) {
						matchingBranches.add(nextOldBranch);
					}
				}
				
				if (!foundPerfectMatch) {
					int index = mostAbstractMatch(matchingBranches);
					//System.out.println("DEBUG: matchingBranches: "+matchingBranches);
					oldBranch = matchingBranches.get(index);
				} else {
					int index = mostAbstractMatch(matchingPerfectBranches);
					oldBranch = matchingPerfectBranches.get(index);
				}
				
				// 2010.02.08
				ArrayList<String> updatedBranch = updateBranchUp(nextRawBranch, oldBranch);
				extendedBranches.add(new WeightBranch(updatedBranch, weight));
				this.LogGenerator.log("Trace : " + key + " " + updatedBranch);
			}
			
			this.LogGenerator.log("} ");
			
			// then compress the updated branches
			//this.mergeSameNodes(extendedBranches, finalStringBranches);
			
			// call the constructSpecialBranchTree method in DataProcessor 
			// to construct a new tree
			this.dataProcessor.constructSpecialBranchTree(futureRoot, extendedBranches);
		}
		
		// move down
		else {
			ArrayList<ArrayList<String>> stringBranches = tree2StringBranches(this.visRoot);
			ArrayList<WeightBranch> extendedBranches = new ArrayList<WeightBranch>();
			
			// for each raw string
			Enumeration<String> keys = this.dataProcessor.rawBranchCount.keys();
			
			this.LogGenerator.log("{ ");
			while(keys.hasMoreElements()) {
				String key = keys.nextElement();
				ArrayList<String> nextRawBranch = this.dataProcessor.rawBranches.get(key);
				int weight = this.dataProcessor.rawBranchCount.get(key);
				
				// find the matching existing branch
				boolean foundPerfectMatch = false;
				ArrayList<ArrayList<String>> matchingBranches = new ArrayList<ArrayList<String>>();
				ArrayList<ArrayList<String>> matchingPerfectBranches = new ArrayList<ArrayList<String>>();
				
				ArrayList<String> oldBranch = null;
				for (int j = 0; j < stringBranches.size(); j++) {
					ArrayList<String> nextOldBranch = stringBranches.get(j);
					
					int result = branchPatternMatches(nextOldBranch, nextRawBranch);
					
					if (result == 1) {
						matchingPerfectBranches.add(nextOldBranch);
						foundPerfectMatch = true;
					}
					if (result == 2) {
						matchingBranches.add(nextOldBranch);
					}
				}
				if (!foundPerfectMatch) {
					int index = mostAbstractMatch(matchingBranches);
					oldBranch = matchingBranches.get(index);
				} else {
					int index = mostAbstractMatch(matchingPerfectBranches);
					oldBranch = matchingPerfectBranches.get(index);
				}
				
				//System.out.println("============================");
				ArrayList<String> updatedBranch = updateBranchDown(nextRawBranch, oldBranch);
				extendedBranches.add(new WeightBranch(updatedBranch, weight));
				
				this.LogGenerator.log("Trace : " + key + " " + updatedBranch);
			}
			this.LogGenerator.log("} ");
			
			this.dataProcessor.constructSpecialBranchTree(futureRoot, extendedBranches);
		}
		
		this.animationRecords = this.displayMechanism.computeMoves(this.visRoot, this.futureRoot, moveUp);
		startAnimation();
	}
	
	public ArrayList<String> updateBranchUp(ArrayList<String> rawBranch, ArrayList<String> oldBranch) {
		int rawBranchIndex  = rawBranch.size() - 1;
		int treeBranchIndex = oldBranch.size() - 1;
		ArrayList<String> updatedBranch = new ArrayList<String>();
		while (rawBranchIndex >= 0 && treeBranchIndex >= 0) {
			String nextFromTreeBranch = oldBranch.get(treeBranchIndex);
			String nextFromRawBranch = rawBranch.get(rawBranchIndex);
			//System.out.println(nextFromTreeBranch + " " + nextFromRawBranch);
			if (specialMatches(nextFromTreeBranch, nextFromRawBranch)) {
				// truncate to the level - 1 of node in old branch
				StringTokenizer st = new StringTokenizer(nextFromTreeBranch, ".");
				
				int currentLevel = st.countTokens();
				int updatedLevel = currentLevel - 1; // level to keep
				
				st = new StringTokenizer(nextFromRawBranch, ".");
				int totalLevel = st.countTokens();
				
				int levelToTruncate = totalLevel - updatedLevel;
				
				String parent = null;
				if (updatedBranch.size() > 0) {
					parent = updatedBranch.get(updatedBranch.size() - 1);
				}
				
				if (currentLevel > 1) {
					StringTokenizer parentSt = null;
					int parentLevel = -1;
					String parentHead = null;
					if (parent != null) {
						parentSt = new StringTokenizer(parent, ".");
						parentLevel = parentSt.countTokens();
						parentHead = this.dataProcessor.truncateLowLevels(parent, 
								0, 1);
					}
					String updatedString = 
						this.dataProcessor.truncateLowLevels(nextFromRawBranch, 
								0, levelToTruncate);
					
					// System.out.println("parentLevel:"+parentLevel+" currentLevel:"+currentLevel+" parentHead:"+parentHead+" updatedString:"+updatedString);
					
					// [CHANGE] FT26Apr11: Trying to correcting bug with wholesame extension / compaction
					// preventing child to compact beyond parent.
					//					if (parent != null && parentLevel >= currentLevel && parentHead.equals(updatedString)) {
					if (parent != null && parentLevel >= currentLevel && parentHead.startsWith(updatedString)) {
						updatedBranch.add(nextFromTreeBranch);
					}
					else {
						updatedBranch.add(updatedString);
					}
				}
				else {
					updatedBranch.add(nextFromTreeBranch);
				}
				
				rawBranchIndex--;
			} else {
				treeBranchIndex--;
			}
		}
		
		String nextParent = "";
		ArrayList<String> finalBranch = new ArrayList<String>();
		
		for (int j = 0; j < updatedBranch.size(); j++) {
			if (j == 0) {
				nextParent = updatedBranch.get(j);
				finalBranch.add(nextParent);
			} else {
				String child = updatedBranch.get(j);
				String parent = nextParent;
				
				if (!this.specialMatches(parent, child)) {
					finalBranch.add(updatedBranch.get(j));
					nextParent = child;
				}
			}
		}
		
		return finalBranch;
	}
	
	public ArrayList<String> updateBranchDown(ArrayList<String> rawBranch, ArrayList<String> oldBranch) {
		// nextRawBranch  oldBranch
		int rawBranchIndex = rawBranch.size() - 1;
		int treeBranchIndex = oldBranch.size() - 1;
		ArrayList<String> updatedBranch = new ArrayList<String>();
		while (rawBranchIndex >= 0 && treeBranchIndex >= 0) {
			String nextFromTreeBranch = oldBranch.get(treeBranchIndex);
			String nextFromRawBranch  = rawBranch.get(rawBranchIndex);
			//System.out.println(nextFromTreeBranch + " " + nextFromRawBranch);
			if (specialMatches(nextFromTreeBranch, nextFromRawBranch)) {
				// truncate to the level + 1 of node in old branch
				StringTokenizer st = new StringTokenizer(nextFromTreeBranch, ".");
				int currentLevel = st.countTokens();
				int updatedLevel = currentLevel + 1; // level to keep
				st = new StringTokenizer(nextFromRawBranch, ".");
				int totalLevel = st.countTokens();
				int levelToTruncate = totalLevel - updatedLevel;
				
				String parent = null;
				if (updatedBranch.size() > 0) {
					parent = updatedBranch.get(updatedBranch.size() - 1);
				}
				
				if (levelToTruncate >= 0) {
					String updatedString = 
						this.dataProcessor.truncateLowLevels(nextFromRawBranch, 0, levelToTruncate);
					if (parent == null || !parent.equals(updatedString)) {
						updatedBranch.add(updatedString);
					}
				}
				else {
					if (parent == null || !parent.equals(nextFromTreeBranch)) {
						updatedBranch.add(nextFromTreeBranch);
					}
				}
				
				rawBranchIndex--;
			} else {
				treeBranchIndex--;
			}
		}
		
		return updatedBranch;
	}
	
	public ArrayList<String> updateBranchWithoutChange(ArrayList<String> rawBranch, ArrayList<String> oldBranch) {
		int rawBranchIndex = rawBranch.size() - 1;
		int treeBranchIndex = oldBranch.size() - 1;
		ArrayList<String> updatedBranch = new ArrayList<String>();
		while (rawBranchIndex >= 0 && treeBranchIndex >= 0) {
			String nextFromTreeBranch = oldBranch.get(treeBranchIndex);
			String nextFromRawBranch = rawBranch.get(rawBranchIndex);
			//System.out.println(nextFromTreeBranch + " " + nextFromRawBranch);
			if (specialMatches(nextFromTreeBranch, nextFromRawBranch)) {
				String parent = null;
				if (updatedBranch.size() > 0) {
					parent = updatedBranch.get(updatedBranch.size() - 1);
				}
				
				if ( parent == null || !parent.equals(nextFromTreeBranch)) {
					updatedBranch.add(nextFromTreeBranch);
				}
				rawBranchIndex--;
			} else {
				treeBranchIndex--;
			}
		}
		// System.out.println(rawBranch + "\n" + updatedBranch + "\n\n");
		return updatedBranch;
	}
	
	public boolean isIdenticalBranch (ArrayList<VisNode> a, ArrayList<VisNode> b) {
		boolean result = true;
		if (a.size() == b.size()) {
			for (int i = 0; i < a.size(); i++) {
				if (!a.get(i).packageName.equals(b.get(i).packageName)) {
					result = false;
				}
			}
			return result;
		}
		else {return false;}
	}

	static void printUsageAndExit() {
			System.out.println("Usage: profvis <Tool> <Mode> [hprof-file]\n"
					+ "1. <Tool>" + " options: \n ProfilingVis \n" + " TreeTable \n" 
					+ "2. <Mode>" + " options: \n replay \n"       + " experiment \n" + " debug \n");
			System.exit(0);
	} // EndMethod printUsageAndExit
	
	static String getNameOfLogFile(String USERNAME, String toolName) {
		Calendar calendar = new GregorianCalendar();
		return
			USERNAME + "_" + toolName + "_"
			+  calendar.get(Calendar.YEAR  )      + "_" 
			+ (calendar.get(Calendar.MONTH ) + 1) + "_"
			+  calendar.get(Calendar.DATE  )      + "_"
			+  calendar.get(Calendar.HOUR  )      + "_"
			+  calendar.get(Calendar.MINUTE)      + "_"
			+  calendar.get(Calendar.SECOND);
	} // EndMethod getNameOfLogFile

	public static OptionSet options ;

	public static void main(String args[]) {


		options = OptionParser.parseOption(args);

		int visType     = options.getTool().equals("treetable")?1:0;
		replayMode      = options.getMode().equals("replay");
		experimentMode  = options.getMode().equals("experiment");
		debugMode       = options.getMode().equals("debug");
		altFormat       = options.isAltformat();
		span            = options.getSpan();

		ProfilingFile   = options.getHprofFile();

		if (visType == 0) {
			
			RobotReplay replay = null; 
			
			if (replayMode) {
				replay = new RobotReplay();
			}
			
			if (experimentMode) {
				USERNAME = JOptionPane.showInputDialog("Please enter your name: ");
				if (USERNAME != null && !USERNAME.equals("")) {
					
					Calendar calendar = new GregorianCalendar();
					USERNAME = getNameOfLogFile( USERNAME, "ProfilingVis" );					
					LogGenerator = new LogGenerator("../log/" + USERNAME + ".txt", experimentMode);
					
					//  0 for yes and 1 for no
					Tutorial = JOptionPane.showConfirmDialog(null, 
							"Would you like to read the tutorial",
							"Tutorial", JOptionPane.YES_NO_OPTION);
					PApplet.main(new String[] { "--present", "src.GraphicPanel" });
				} 
				else {
					System.exit(0);
				}
			}
		
			if (debugMode){
				// For some weird buggy reason I need this on a Mac for other
				// dialogs to work reliably
				JOptionPane.showMessageDialog(null,"Welcome to ProfiViz\nRunning in debug mode");
				LogGenerator = new LogGenerator("../log/last.txt", false);
				if (options.isWindow()) {
					JFrame frame = new FrancoisFrame(800,500);
					frame.setVisible(true);
				} else {
					PApplet.main(new String[] { "--present", "src.GraphicPanel" });
				}
			}
		
			if (replayMode) {
				String filename = JOptionPane.showInputDialog("Please input a replay file: ");
				LogGenerator = new LogGenerator("../log/last.txt", false);
				PApplet.main(new String[] { "--present", "src.GraphicPanel" });
				replay.replay("log/" + filename + ".txt");
			}	
		}
	
	
		if (visType == 1) {
		
			if (replayMode) {
				System.out.println("Replay Mode is not provided for TreeTable");
			}
			
			if (experimentMode) {
				USERNAME = JOptionPane.showInputDialog("Please enter your name: ");
				if (USERNAME != null && !USERNAME.equals("")) {
					USERNAME = getNameOfLogFile( USERNAME, "TreeTable" );										
					File file = new File("log/" + USERNAME + ".txt");
					try {
						file.createNewFile();
					} catch (IOException e) {
						e.printStackTrace();
					}
					JFrame myFrame = new TreeTable() ;
					myFrame.setVisible(true);	
				}
				else {
					System.exit(0);
				}
			}
					
			if (debugMode) {
				JFrame myFrame = new TreeTable() ;
				myFrame.setVisible(true);	
			}
		}
	}
}

class FrancoisFrame extends JFrame {

	GraphicPanel myGraphicPan ;

	public FrancoisFrame(int width, int height) {
		super("ProfiVis");

		setLayout(new BorderLayout());
		setSize(width, height); 

		myGraphicPan = new GraphicPanel(width, height, 16, this);
		//myGraphicPan.BACKGROUND_COLOUR = 200;
		//myGraphicPan.textColour = 20;
		//CircleLayout.ORIGIN_FONT_SIZE = 14;
		add(myGraphicPan, BorderLayout.CENTER);
		
		// important to call this whenever embedding a PApplet.
		// It ensures that the animation thread is started and
		// that other internal variables are properly set.
		myGraphicPan.init();
	} // EndConstructor FrancoisFrame

	public void paint(Graphics g) {
		super.paint(g);
		if (myGraphicPan!=null) myGraphicPan.redraw();
	}

} // EndClass FrancoisFrame

class GraphFormat {
	public int ORIGIN_FONT_SIZE     = 10;
 	public int ORIGIN_H_SPACE       = 110;
	public int ORIGIN_V_SPACE       = 80;
	public int ORIGIN_WIDTH         = 100;
	public int ORIGIN_HEIGHT        = 60;
	public int ORIGIN_HALF_WIDTH    = ORIGIN_WIDTH / 2;
	public int ORIGIN_HALF_HEIGHT   = ORIGIN_HEIGHT / 2;
	public int ORIGIN_RECT_HEIGHT   = 13;
	public int ORIGIN_RECT_WIDTH    = 3;
	public int ORIGIN_PARENT_WIDTH  = 15;
	public int ORIGIN_PARENT_HEIGHT = 15;
	public int ORIGIN_CIRCLE_ARROW_RADIUS = 10;

	public float ORIGIN_RADIUS = 120;

	public int BACKGROUND_COLOUR = 60;

	public double START_ANGLE = 170 ; // for circular layout
	public double END_ANGLE   = 370 ;

	public DataProcessor myDataProcessor ;
	public GraphicPanel  myGraphicPanel  ;
	public boolean doDrawParentColour = true ;
	public boolean doDrawIcon = true;

	public GraphFormat(GraphicPanel graphicPanel, DataProcessor dataProcessor,int span) {
		START_ANGLE = 270.0 - span/2.0;
		END_ANGLE   = 270.0 + span/2.0;
		myGraphicPanel  = graphicPanel ;
		myDataProcessor = dataProcessor;
	} // EndConstructor

	public void colorAllocation() {
		myDataProcessor.colorAllocation(myDataProcessor.packageTree, 0, 255, 0, 255, 0, 255);
	} // EndMethod colorAllocation


	public void nodeFormat(int[] nodeColor, boolean dimmed) {
		// float dimFactor = dimmed?1.5f:1;
		// int[] nodeColor = new int[3];
		// for(int i = 0; i<3; i++) { nodeColor[i] = (int)((origColor[i]-50)/dimFactor) + 50; }
		myGraphicPanel.fill  (nodeColor[0]  , nodeColor[1]  , nodeColor[2]); 
		myGraphicPanel.strokeWeight (2);
		myGraphicPanel.stroke(nodeColor[0]/2, nodeColor[1]/2, nodeColor[2]/2); 
	} // EndMethod nodeFormat

	public void textFormat(int[] nodeColor, boolean zoomText, boolean dimmed) {
		
		if (zoomText) {
			myGraphicPanel.fill(255, 255, 255);
			return;
		}

		if (nodeColor[0] + nodeColor[1] + nodeColor[2] < 325) {
			myGraphicPanel.fill(255, 255, 255, /*dimmed?150:200*/ 200);
		}
		else {
			myGraphicPanel.fill(0, 0, 0/*, dimmed?150:255*/);
		}
	} // EndMethod textFormat

	public int getArrowColour(String packageName) {

		return (packageName.equals("Null")
						? myGraphicPanel.color(150,0,150,192)
						: myGraphicPanel.color(200));

	} // EndMethod getArrowColour

	public int getArrowSize(int defaultArrowSize, double childScaleVar) {
		int newCircleArrowRadius = (int)(defaultArrowSize/ childScaleVar);
		if (newCircleArrowRadius <= 8) {
			newCircleArrowRadius = 8;
		}
		return newCircleArrowRadius;
	} // EndMethod getArrowSize

} // EndGraphFormat

class AlternativeGraphFormat extends GraphFormat {

	public AlternativeGraphFormat(GraphicPanel graphicPanel, DataProcessor dataProcessor,int span) {
		super(graphicPanel, dataProcessor,span);
		BACKGROUND_COLOUR = 255;
		ORIGIN_RADIUS     = 80;
		ORIGIN_FONT_SIZE  = 18;
		doDrawParentColour = false ;
		doDrawIcon         = false ;
	} // EndConstructor

	public void colorAllocation() {
		myDataProcessor.colorAllocation(myDataProcessor.packageTree, 255, 0, 0, 255, 0, 255);
	} // EndMethod colorAllocation


	public void nodeFormat(int[] nodeColor, boolean dimmed) {
		myGraphicPanel.stroke(nodeColor[0], nodeColor[1], nodeColor[2],200);
		myGraphicPanel.strokeWeight((float)(4*myGraphicPanel.graphicState.SCALE)); 
		myGraphicPanel.fill(255); 
	} // EndMethod nodeFormat

	public void textFormat(int[] nodeColor, boolean zoomText) {
		
		myGraphicPanel.fill(0);

	} // EndMethod textFormat

	public int getArrowColour(String packageName) {

		return 0;

	} // EndMethod getArrowColour

	public int getArrowSize(int defaultArrowSize, double childScaleVar) {
		int newCircleArrowRadius = (int)(defaultArrowSize);// / childScaleVar);
		if (newCircleArrowRadius<4) {newCircleArrowRadius *=2;}
		else if (newCircleArrowRadius<8) {newCircleArrowRadius = 8;}
		return newCircleArrowRadius;
	} // EndMethod getArrowSize
}

class GraphicPanelState implements Cloneable {
	public double SCALE    = 1.2;
	public double distX    = 0;
	public double distY    = 0;
	public boolean invalid = false ;
	public VisNode nodeUnderMouse = null;
	public int furtherCompression = 0;

	public /*synchronized*/ void  changeScale(double delta) {
		double oldScale = SCALE ;
		this.SCALE += delta;
		this.distX *= SCALE/oldScale;
		this.distY *= SCALE/oldScale;
		//System.out.println("DistY: "+distY);
	} // EndMethod changeScale 

	public GraphicPanelState copy() {
		GraphicPanelState res = null ;
		try {
		 res = (GraphicPanelState)super.clone();
		} catch (Exception e) {
			System.err.println(e);
			e.printStackTrace();
			System.exit(1);
		}
		assert(this.invalid || !this.hasChanged(res));
		return res;
	} // EndMethod copy

	public void invalidate() {
		invalid = true;
	}

	public boolean hasChanged(GraphicPanelState previous) {
		if (previous==null) return true;
		if (previous.invalid || this.invalid) return true;
		boolean res = false;
		try {
			for (Field f: getClass().getFields()) {
				if (f.get(this)!=null)
					res = res || !(f.get(this).equals(f.get(previous)));
				else
					res = res || !(f.get(this)==f.get(previous));
				//System.err.println(f.getName()+":"+f.get(this)+":"+f.get(previous)+":"+res);
			}
		} catch (Exception e) {
			System.err.println(e);
			e.printStackTrace();
			System.exit(1);
		}
		return res;
	} // EndMethod hasChanged
}
