package src;

import processing.core.PConstants;

/**
 * Based on the example in Processing tutorial
 * @author lins6
 * 
 */
public class HScrollBar {

	public GraphicPanel graphicPanel;

	public int swidth, sheight;    // width and height of bar
	public int xpos, ypos;         // x and y position of bar
	public float spos, newspos;    // x position of slider
	public int sposMin, sposMax;   // max and min values of slider
	public int loose;              // how loose/heavy
	public boolean over;           // is the mouse over the slider?
	public boolean locked;
	public float ratio;


	public HScrollBar(GraphicPanel graphicPanel, int xp, int yp, int sw, int sh, int l) {
		this.graphicPanel = graphicPanel;
		swidth = sw;
		sheight = sh;
		int widthtoheight = sw - sh;
		ratio = (float)sw / (float)widthtoheight;
		xpos = xp;
		ypos = yp-sheight/2;
		spos = xpos + swidth/2 - sheight/2;
		newspos = spos;
		sposMin = xpos;
		sposMax = xpos + swidth - sheight;
		loose = l;
	}

	public void update() {
		if(over()) {
			over = true;
		} else {
			over = false;
		}
		
		if(this.graphicPanel.mousePressed && over) {
			locked = true;
		}
		
		if(!this.graphicPanel.mousePressed) {
			locked = false;
		}
		
		if(locked) {
			newspos = constrain(this.graphicPanel.mouseX-sheight/2, sposMin, sposMax);
		}
		
		if(Math.abs(newspos - spos) > 1) {
			spos = spos + (newspos-spos)/loose;
		}
	}
	
	public int constrain(int val, int minv, int maxv) {
		return Math.min(Math.max(val, minv), maxv);
	}

	public boolean over() {
		if(this.graphicPanel.mouseX > xpos && 
		   this.graphicPanel.mouseX < xpos+swidth &&
		   this.graphicPanel.mouseY > ypos && 
		   this.graphicPanel.mouseY < ypos+sheight) {
			return true;
		} else {
			return false;
		}
	}

	public void display() {
		this.graphicPanel.fill(255);
		this.graphicPanel.rectMode(PConstants.CORNER);
		this.graphicPanel.rect(xpos, ypos, swidth, sheight);
		if(over || locked) {
			this.graphicPanel.fill(153, 102, 0);
		} else {
			this.graphicPanel.fill(102, 102, 102);
		}
		
		int graphWidth = this.graphicPanel.rightBorder - this.graphicPanel.leftBorder;
		int slideWidth = this.graphicPanel.SCREEN_WIDTH * swidth / graphWidth;
		
		this.graphicPanel.rect(spos, ypos, slideWidth, sheight);
	}

	float getPos() {
		// Convert spos to be values between
		// 0 and the total width of the scrollbar
		return spos * ratio;
	}
}
