/**
 * OptionParser.java
 * 
 * Time-stamp: <2011-04-28 13:50:09 ftaiani>
 * @author Francois Taiani   <f.taiani@lancaster.ac.uk>
 * $Id$
 * 
 */

package src;

import joptsimple.*;

/**
 * @author Francois Taiani   <f.taiani@lancaster.ac.uk>
 */
public abstract class OptionParser {

  static final private joptsimple.OptionParser parser = new joptsimple.OptionParser();

	static src.OptionSet parseOption(String args[]) {
    parser.accepts( "tool" ).withRequiredArg()
      .describedAs( "profilingvis|treetable")
      .defaultsTo ( "profilingvis" );
    parser.accepts( "mode" ).withRequiredArg()
      .describedAs( "debug|experiment|replay")
      .defaultsTo ( "debug" );
    parser.accepts( "window"   , "windowed mode");
    parser.accepts( "nohint"   , "don't display hints");
    parser.accepts( "altformat", "alternative formatting");
    parser.accepts( "span").withRequiredArg()
      .ofType( Integer.class )
      .describedAs("angle")
      .defaultsTo(240);
    parser.accepts( "precision","approx. factor (%)").withRequiredArg()
      .ofType( Double.class )
      .describedAs("precision")
      .defaultsTo(0.5);
    parser.accepts( "h", "show help" );

    joptsimple.OptionSet rawSet = parser.parse(args);
    if (rawSet.has("h")) printHelp();
    return new src.OptionSet(rawSet);
  } // EndMethod OptionSet

  static void printHelp() {
    System.err.println("Usage: profvis [options] <hproffile>\n");
    try {
      parser.printHelpOn( System.err );
      System.exit(0);
    } catch (Exception e) {
      System.err.println(e);
      System.exit(1);
    }
  }

} // EndClass OptionParser
