package src;

import java.awt.AWTException;
import java.awt.Robot;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.StringTokenizer;

public class RobotReplay {

	public Robot robot;
	
	public RobotReplay() {
		try {
			this.robot = new Robot();
		} catch (AWTException e) {
			e.printStackTrace();
		}
	}
	
	public void replay(String file) {
		FileReader fr = null;
		String nextLine = "";
		long prevTime = -1;
		try {
			fr = new FileReader(file);
			BufferedReader br = new BufferedReader(fr);
			
			while( (nextLine = br.readLine()) != null) {
				prevTime = action(prevTime, nextLine);
			}
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	public static String MOUSE_PRESS   = "MousePress";
	public static String MOUSE_RELEASE = "MouseRelease";
	public static String SCALE_UP      = "ScaleUp";
	public static String SCALE_DOWN    = "ScaleDown";
	public static String EXPAND        = "Expand";
	public static String AGGREGATE     = "Aggregate";
	
	public long action(long prevTime, String command) {
		
		StringTokenizer st = new StringTokenizer(command, ";");
		ArrayList<String> commandList = new ArrayList<String>();
		while (st.hasMoreTokens()) {
			commandList.add(st.nextToken());
		}
		
		long time = -1;
		
		if (command.contains(MOUSE_PRESS)) {
			int x = 0;
			int y = 0;
			for (int i = 0; i < commandList.size(); i++) {
				String nextInfo = commandList.get(i);
				if (nextInfo.contains(MOUSE_PRESS)) {
					StringTokenizer tmpST = new StringTokenizer(nextInfo, ":");
					tmpST.nextToken();
					String coordinatorStr = tmpST.nextToken();
					tmpST = new StringTokenizer(coordinatorStr, "-");
					x = Integer.parseInt(tmpST.nextToken());
					y = Integer.parseInt(tmpST.nextToken());
				}
				if (nextInfo.contains("Time")) {
					StringTokenizer tmpST = new StringTokenizer(nextInfo, ":");
					tmpST.nextToken();
					time = Long.parseLong(tmpST.nextToken());
				}
			}
			
			if (prevTime < 0) {
				robot.delay(1000);
			} else {
				robot.delay((int) (time - prevTime));
			}
			
			robot.mouseMove(x, y);
			robot.mousePress(InputEvent.BUTTON1_MASK);
		}
		
		if (command.contains(MOUSE_RELEASE)) {
			int x = 0;
			int y = 0;
			boolean isLeft = true;
			for (int i = 0; i < commandList.size(); i++) {
				String nextInfo = commandList.get(i);
				if (nextInfo.contains("MouseRelease")) {
					StringTokenizer tmpST = new StringTokenizer(nextInfo, ":");
					tmpST.nextToken();
					String coordinatorStr = tmpST.nextToken();
					tmpST = new StringTokenizer(coordinatorStr, "-");
					x = Integer.parseInt(tmpST.nextToken());
					y = Integer.parseInt(tmpST.nextToken());
				}
				if (nextInfo.contains("LeftClick")) {
					isLeft = true;
				}
				if (nextInfo.contains("RightClick")) {
					isLeft = false;
				}
				if (nextInfo.contains("Time")) {
					StringTokenizer tmpST = new StringTokenizer(nextInfo, ":");
					tmpST.nextToken();
					time = Long.parseLong(tmpST.nextToken());
				}
			}
			
			if (prevTime < 0) {
				robot.delay(1000);
			} else {
				robot.delay((int) (time - prevTime));
			}
			
			robot.mouseMove(x, y);
			if (isLeft) {
				robot.mouseRelease(InputEvent.BUTTON1_MASK);
			} else {
				robot.mouseRelease(InputEvent.BUTTON3_MASK);
			}
		}
		
		if (command.contains(SCALE_UP)) {
			for (int i = 0; i < commandList.size(); i++) {
				String nextInfo = commandList.get(i);
				if (nextInfo.contains("Time")) {
					StringTokenizer tmpST = new StringTokenizer(nextInfo, ":");
					tmpST.nextToken();
					time = Long.parseLong(tmpST.nextToken());
				}
			}
			
			if (prevTime < 0) {
				robot.delay(1000);
			} else {
				robot.delay((int) (time - prevTime));
			}
			
			robot.keyPress(KeyEvent.VK_EQUALS);
		}
		
		if (command.contains(SCALE_DOWN)) {
			for (int i = 0; i < commandList.size(); i++) {
				String nextInfo = commandList.get(i);
				if (nextInfo.contains("Time")) {
					StringTokenizer tmpST = new StringTokenizer(nextInfo, ":");
					tmpST.nextToken();
					time = Long.parseLong(tmpST.nextToken());
				}
			}
			
			if (prevTime < 0) {
				robot.delay(1000);
			} else {
				robot.delay((int) (time - prevTime));
			}
			
			robot.keyPress(KeyEvent.VK_MINUS);
		}
		
		if (command.contains(EXPAND)) {
			for (int i = 0; i < commandList.size(); i++) {
				String nextInfo = commandList.get(i);
				if (nextInfo.contains("Time")) {
					StringTokenizer tmpST = new StringTokenizer(nextInfo, ":");
					tmpST.nextToken();
					time = Long.parseLong(tmpST.nextToken());
				}
			}
			
			if (prevTime < 0) {
				robot.delay(1000);
			} else {
				robot.delay((int) (time - prevTime));
			}
			
			robot.keyPress(KeyEvent.VK_RIGHT);
		}
		
		if (command.contains(AGGREGATE)) {
			for (int i = 0; i < commandList.size(); i++) {
				String nextInfo = commandList.get(i);
				if (nextInfo.contains("Time")) {
					StringTokenizer tmpST = new StringTokenizer(nextInfo, ":");
					tmpST.nextToken();
					time = Long.parseLong(tmpST.nextToken());
				}
			}
			
			if (prevTime < 0) {
				robot.delay(1000);
			} else {
				robot.delay((int) (time - prevTime));
			}
			
			robot.keyPress(KeyEvent.VK_LEFT);
		}
		return time;
	}
}
