/**
 * The tree structure modeling
 * 
 * @author Shen Lin
 */

package src;

import java.util.*;

public class VisNode {
	
	float x;
	float y;
  float oldx;
  float oldy;

  boolean highlighted = false ;

	int widthFactor;
	boolean expanded = true;
	
	boolean visible = true;
	boolean hide = true;
	
	VisNode parent;
    //	ArrayList<VisNode> children;

	SortedSet<VisNode> children;

	
	int profile;
	String packageName;
	
	public VisNode(String name, int profile) {
		this.packageName = name;
		this.children = new TreeSet<VisNode>(new VisNodeComparator());
		this.profile = profile;
	}
	
	/******************
	 * Visualisation methods
	 ******************/
	
	public VisNode copy() {
		VisNode futureRoot = new VisNode("Null", -1);
		futureRoot.x = this.x;
		futureRoot.y = this.y;
		futureRoot.oldx = this.oldx;
		futureRoot.oldy = this.oldy;
		recursiveCopy(this, futureRoot);
		return futureRoot;
	}
	
	public void recursiveCopy(VisNode node, VisNode futureNode) {
		if (node.children != null && node.children.size() > 0) {
                    //			for (int i = 0; i < node.children.size(); i++) {
                    for(VisNode nextChild : node.children) {
                        //				VisNode nextChild = node.children.get(i);
                        VisNode nextFutureChild = new VisNode("Null", -1);
                        nextFutureChild.x = nextChild.x;
                        nextFutureChild.y = nextChild.y;
                        nextFutureChild.oldx = nextChild.oldx;
                        nextFutureChild.oldy = nextChild.oldy;
                        nextFutureChild.packageName = nextChild.packageName;
                        nextFutureChild.profile = nextChild.profile;
                        futureNode.addChild(nextFutureChild);
                        nextFutureChild.setParent(futureNode);
                        recursiveCopy(nextChild, nextFutureChild);
                    }
		}
	}
	
	public void setExpanded(boolean expanded) {
		this.expanded = expanded;
	}
	
	public boolean getExpanded() {
		return this.expanded;
	}
	
	public void setLocation(float x,float y) {
		this.x = x;
		this.y = y;
		this.oldx = x;
		this.oldy = y;
	}
	
	/*************************
	 * Data Structure methods
	 *************************/
	
	public String toString() {
		return this.packageName;
	}
	
	public void setParent(VisNode parent) {
		this.parent = parent;
	}
	
	public VisNode getParent() {
		return this.parent;
	}
	
	public void addChild(VisNode child) {
		this.children.add(child);
		child.setParent(this);
	}
	
	public SortedSet<VisNode> getChildren() {
		return children;
	}
}

class VisNodeComparator implements Comparator<VisNode> {
    public int compare(VisNode o1,VisNode o2) {
        int res = o1.packageName.compareTo(o2.packageName);
        return res;
    }
}
