/* -*-  Mode:Java; c-basic-offset:2; tab-width:1; indent-tabs-mode:nil; -*- */

package treetable;

/**
 * Original code by Geertjan Wielenga, published on 2008/06/04
 * Further modifications (c) Francois Taiani / Lancaster University 2010
 */

import org.netbeans.swing.outline.*;

import src.GraphicPanel;


import javax.swing.*;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.event.TreeExpansionEvent;
import javax.swing.event.TreeExpansionListener;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.table.*;
import javax.swing.tree.*;
import java.awt.*;
import java.io.*;
import java.text.DecimalFormat;
import java.util.*;


public class TreeTable extends JFrame implements ListSelectionListener, TreeExpansionListener {
	
  private Outline outline;
  TreeTableLogGenerator logGenerator;
	
  public TreeTable() { 
    logGenerator = new TreeTableLogGenerator("log/" + GraphicPanel.USERNAME + ".txt", GraphicPanel.ENABLE_LOG);
			
    setDefaultCloseOperation(EXIT_ON_CLOSE);
		
    getContentPane().setLayout(new BorderLayout());
		
    VisNode       root          = new VisNode("Root", -1);
    DataProcessor dataProcessor = new DataProcessor(GraphicPanel.ProfilingFile, root);	
    TreeModel     treeModel     = new FileTreeModel(root);
    RowModel      fileRowModel  = new FileRowModel(dataProcessor);
    OutlineModel mdl = DefaultOutlineModel.createOutlineModel
      (treeModel, fileRowModel, true);
		
    TreePathSupport treePathSupport = mdl.getTreePathSupport();
    treePathSupport.addTreeExpansionListener(this);
		
    outline = new Outline();
    outline.setRootVisible(false);
    outline.setModel(mdl);
    outline.setRenderDataProvider(new RenderData(dataProcessor));
    outline.setColumnSorted(3, false, 1);
    outline.setShowHorizontalLines(true);
    outline.setShowVerticalLines(false);

    JTable outlineTable = (JTable)outline;
    ListSelectionModel listSelectionModel = outlineTable.getSelectionModel();
    listSelectionModel.addListSelectionListener(this);
		
    TableColumn column = outlineTable.getColumnModel().getColumn(3);
    column.setCellRenderer(new GraphicBar(dataProcessor));

    DefaultTableCellRenderer renderer = new DefaultTableCellRenderer();
    renderer.setHorizontalAlignment(JLabel.RIGHT);
    outlineTable.getColumnModel().getColumn(2).setCellRenderer(renderer);
  
    outline.setGridColor(Color.lightGray);
		
    getContentPane().add(new JScrollPane(outline),BorderLayout.CENTER);
    setBounds(20, 20, 800, 600);
		
  } // EndConstructor NewJFrame

  public void valueChanged(ListSelectionEvent lse) {
    // do nothing
  }

  public void treeExpanded(TreeExpansionEvent tee) {
    //System.out.println("Expand " + tee.getPath());
    logGenerator.log("Expand " + tee.getPath());
  }

  public void treeCollapsed(TreeExpansionEvent tee) {
    //System.out.println("Collapse " + tee.getPath());
    logGenerator.log("Collapse " + tee.getPath());
  }

} // EndClass NewJFrame



class DataProcessor {
	
  String[] strings;
  Hashtable<String, ArrayList<String>> rawBranches;
  Hashtable<String, Integer> rawBranchCount = new Hashtable<String, Integer>();
  VisNode root;
  int Total;
  float maxPercent = 0;
  int maxCount ;
  int totalNodes = 0;

  DataProcessor(String file, VisNode root) {
    this.root = root;
    strings = loadStrings(file);
    rawBranches = rawFileProcess(strings);	
    countProcess();
		
    constructTree();
  }
	
  void countProcess() {
    boolean totalReached = false;
		
    for (int i = 0; i < strings.length; i++) {
      if (strings[i].contains("CPU SAMPLES BEGIN")) {
        StringTokenizer st = new StringTokenizer(strings[i]);
        st.nextToken(); st.nextToken(); st.nextToken();
        st.nextToken(); st.nextToken();
        String total = st.nextToken();
        total = total.substring(0, total.length() - 1);
        this.Total = Integer.parseInt(total);
        System.out.println(Total);
        totalReached = true;
      }
			
      if (totalReached && !strings[i].contains("CPU") && !strings[i].contains("rank")) {
        StringTokenizer st = new StringTokenizer(strings[i]);
        st.nextToken(); st.nextToken(); st.nextToken();
				
        String countStr = st.nextToken(); 
        String traceStr = st.nextToken();
				
        int count = Integer.parseInt(countStr);
        rawBranchCount.put(traceStr, count);
      }
    }
  }
	
  void constructTree() {

    int maxCount = 0 ;

    System.out.println(this.rawBranches.size() + " " + this.rawBranchCount.size());
		
    Enumeration<String> keys = rawBranches.keys();
		
    while (keys.hasMoreElements()) {
      String nextKey = keys.nextElement();
      ArrayList<String> branch = rawBranches.get(nextKey);
			
      VisNode currentNode = root;
			
      for (int j = branch.size() - 1; j >= 0; j--) {
        String nextPackage = branch.get(j);
        boolean contains = false;
        VisNode nextChild = null;
        String nextName = "";
        int nextProfile = -1;
				
        if (this.rawBranchCount.get(nextKey)==null) {
          continue; // FT08Apr10: This trace has no count so skip
        }
        int count = this.rawBranchCount.get(nextKey);
				
        for (int k = 0; k < currentNode.children.size(); k++) {
          nextChild = currentNode.children.get(k);
          nextName = nextChild.packageName;
          nextProfile = nextChild.profile;
					
          if (nextName.equals(nextPackage)) {
            contains = true; 
            break;
          }
        } // EndFor k

        int newCount ; // Only use to compute maxCount

        if (contains) {
          nextProfile += count;
          newCount = nextProfile;
          nextChild.profile = nextProfile;
          currentNode = nextChild;
        }
        else {
          VisNode newNode = new VisNode(nextPackage, count);
          totalNodes++;
          newCount = nextProfile;
          currentNode.children.add(newNode);
          newNode.parent = currentNode;
          currentNode = newNode;
        } // EnfIfElse contains

        if (newCount > maxCount) { maxCount = newCount ; }

      } // EndFor j

    } // EndWhile (keys.hasMoreElements())

    maxPercent = (float)maxCount / Total ;
    System.out.println("totalNodes:"+totalNodes);

  } // EndMethod constructTree()
	
  void compareAll(ArrayList<String> a, Hashtable<ArrayList<String>, Integer> table) {
    Enumeration<ArrayList<String>> keys = table.keys();
		
    boolean equal = false;
    ArrayList<String> b = null;
		
    while(keys.hasMoreElements()) {
      b = keys.nextElement();
      if ( compare (a, b) ) {
        equal = true;
        break;
      }
    }
			
    if (equal) {
      Integer count = table.get(b);
      table.put(a, count++);
    } else { table.put(a, 1); }
  }
	
  boolean compare(ArrayList<String> a, ArrayList<String> b) {
    boolean result = true;
    if (a.size() == b.size()) {
      for (int i = 0; i < a.size(); i++) {
        if ( !a.get(i).equals(b.get(i)) ) {
          result = false;
          break;
        }
      }
			
    } else { return false; }
		
    return result;
  }
	
  public Hashtable<String, ArrayList<String>> rawFileProcess(String[] theFile) {
    boolean traceSectionStart = false;
    int initIndex = -1;
    String traceID = "";
    Hashtable<String, ArrayList<String>> processedFile = 
      new Hashtable<String, ArrayList<String>>();
		
    for (int i = 0; i < theFile.length; i++) {
      if (theFile[i].contains("--------")) {
        traceSectionStart = true;
      }

      if (traceSectionStart) {
        if (theFile[i].contains("TRACE") || theFile[i].contains("CPU SAMPLES BEGIN")) {
					
          if (initIndex == -1) {
            initIndex = i + 1;
						
            StringTokenizer stID = new StringTokenizer(theFile[i]);
            stID.nextToken();
            String idStr = stID.nextToken();
            idStr = idStr.substring(0, idStr.length() - 1);
            //System.out.println("whats id: " + idStr);
            traceID = idStr;
						
          } else {

            ArrayList<String> branch = new ArrayList<String>();
            boolean isEmptyBranch = false;
            for (int j = initIndex; j < i; j++) {
              StringTokenizer st = new StringTokenizer(theFile[j], "(");
              String tmp = st.nextToken();
              if (tmp.contains("<empty>")) {
                isEmptyBranch = true;
                break;
              }
              else {
                branch.add(tmp);
              }
            }
						
            if (!isEmptyBranch) {
              processedFile.put(traceID, branch);
            }
						
            StringTokenizer stID = new StringTokenizer(theFile[i]);
            stID.nextToken();
            String idStr = stID.nextToken();
            if (!idStr.contains("SAMPLE")) {
              idStr = idStr.substring(0, idStr.length() - 1);
              //System.out.println("whats id: " + idStr);
              traceID = idStr;
            }
            initIndex = i + 1;
          }
        }
      }
    }
    return processedFile;
  }
	
  public String[] loadStrings(String file) {
    ArrayList<String> lines = new ArrayList<String>();
    try {
      BufferedReader br = new BufferedReader(new FileReader(file));
      String tmp = "";
      while((tmp = br.readLine()) != null) {
        lines.add(tmp);
      }
    } catch (Exception e) {
      e.printStackTrace();
    }
		
    int size = lines.size();
    String[] results = new String[size];
    for (int i = 0; i < size; i++) {
      results[i] = lines.get(i);
    }
    return results;
  }
}


class VisNode {
  public VisNode parent;
  public ArrayList<VisNode> children;
  
  public int profile;
  public String packageName;
	
  public VisNode(String name, int profile) {
    this.packageName = name;
    this.children = new ArrayList<VisNode>();
    this.profile = profile;
  }
	
  public String toString() {
    return packageName;
  }
}


class FileTreeModel implements TreeModel {
	
  private VisNode root;
	
  public FileTreeModel(VisNode root) {
    this.root = root;	
  }
	
  public void valueChanged(TreeSelectionEvent arg0) {
    // do nothing
  }
	
  public void addTreeModelListener(javax.swing.event.TreeModelListener l) {
    //do nothing
  }
  public Object getChild(Object parent, int index) {
    VisNode v = (VisNode) parent;
    return v.children.get(index);
  }
  public int getChildCount(Object parent) {
    VisNode v = (VisNode) parent;
    return v.children.size();
  }
  public int getIndexOfChild(Object parent, Object child) {
    VisNode par = (VisNode) parent;
    VisNode ch = (VisNode) child;
    ArrayList<VisNode> children = par.children;
    int index = -1;
    for (int i = 0; i < children.size(); i++) {
      String childrenPackageName = children.get(i).packageName;
      String childPackageName = ch.packageName;
      if (childrenPackageName.equals(childPackageName)) {
        index = i;
        break;
      }
    }
    return index;
  }
  public Object getRoot() {
    return root;
  }
  public boolean isLeaf(Object node) {
    VisNode v = (VisNode) node;
    return v.children.size() == 0;
  }
  public void removeTreeModelListener(javax.swing.event.TreeModelListener l) {
    //do nothing
  }
  public void valueForPathChanged(javax.swing.tree.TreePath path, Object newValue) {
    //do nothing
  }
}

class FileRowModel implements RowModel {
	
  DataProcessor dataProcessor;
	
  public FileRowModel(DataProcessor dataProcessor) {
    this.dataProcessor = dataProcessor;
  }
	
  public Class getColumnClass(int column) {
    switch (column) {
    case 0:
      return Integer.class;
    case 1:
      return String.class;	
    case 2:
      return GraphicBar.class;	
    default:
      assert false;
    }
    return null;
  }
  public int getColumnCount() {
    return 3;
  }
  public String getColumnName(int column) {
    switch (column) {
    case 0:
      return "Execution Frequency";
    case 1:
      return "Percentage";	
    case 2:
      return "Percentage";	
    default:
      assert false;
    }
    return null;
  }
  public Object getValueFor(Object node, int column) {
    VisNode v = (VisNode) node;
    float result = ((float)v.profile / (float)dataProcessor.Total) * 100;
    switch (column) {
    case 0:
      return v.profile;
    case 1:
      DecimalFormat   fnum   =   new   DecimalFormat("##0.00");  
      String dd = fnum.format(result);       
      return dd + "%";
    case 2:
      return result; 
    default:
      assert false;
    }
    return null;
  }
  public boolean isCellEditable(Object node, int column) {
    return false;
  }
  public void setValueFor(Object node, int column, Object value) {
    //do nothing for now
  }
} // EndClass FileRowModel

class RenderData implements RenderDataProvider {
	
  DataProcessor dataProcessor;

  public RenderData(DataProcessor dataProcessor) {
    this.dataProcessor = dataProcessor;
  }

  public java.awt.Color getBackground(Object o) {
    return null;
  }
  public String getDisplayName(Object o) {
    return ((VisNode) o).packageName;
  }
  public java.awt.Color getForeground(Object o) {
    // return UIManager.getColor("controlShadow");
    return null;
  }
	
  public javax.swing.Icon getIcon(Object o) {
    // return new javax.swing.plaf.metal.MetalIconFactory.TreeFolderIcon();
    // return null;
    VisNode node = ((VisNode) o);
    if (node.children.size() == 0) {
      return new javax.swing.plaf.metal.MetalIconFactory.TreeLeafIcon();
    } else {
      return new javax.swing.plaf.metal.MetalIconFactory.TreeFolderIcon();
    }
  }

  public String getTooltipText(Object o) {
    VisNode v = (VisNode) o;
    float result = ((float)v.profile / (float)dataProcessor.Total) * 100;
    DecimalFormat   fnum   =   new   DecimalFormat("##0.00");  
    String dd = fnum.format(result);       

    return v.packageName + " (" + dd + "%)";
  }
  public boolean isHtmlDisplayName(Object o) {
    return false;
  }
} // EndClass RenderData
